# -*- coding: utf-8 -*-

import audioread
import chromaprint
import hashlib
import musicbrainzngs
import mutagen
import os
import urllib
from subprocess import Popen, PIPE
from rusty.models import Model, Manager, QuerySet
from rusty.http import Response
from rusty.jason import json_decode
from rusty.terminal import Colors
from rusty.workers import Worker
from rusty.location import Country
from rusty.authentication import User, LastFmToken
from rusty.views import View
from rusty.forms import Form
from rusty.files import File, Directory, DirectoryIndex, FileField
from rusty.streams import InputStream
from rusty.fields import ForeignKey, UUIDField, IntegerField, StringField, \
    FloatField, YearField, BooleanField, TagsField, URLField, MimetypeField, \
    ListField, DateTimeField, SlugField, URNField, MD5Field

musicbrainzngs.set_useragent("Rusty web framework", "http://bitbucket.org/lauri.vosandi/rusty-web-framework")

"""
Rusty audio module streams
"""

class TranscodeStream(InputStream):
    def __init__(self, mixed, transcode_profile, output_stream=None):
        if isinstance(mixed, File):
            mixed = mixed.get_read_stream()
        assert isinstance(mixed, InputStream), "Don't know how to transcode %s" % mixed
        self.transcode_profile = transcode_profile
        pipe_command = "/usr/bin/avconv -loglevel fatal -i - %s -y -" % self.transcode_profile.options
        proc = Popen(pipe_command.split(), stdin=mixed.fileno(), stdout=output_stream or PIPE, stderr=None)
        self._handle = proc.stdout
        self.mimetype = transcode_profile.mimetype
        self.filename = mixed.filename +  ".mp3"
        
    def __repr__(self):
        return "<TranscodeStream to %s>" % self.transcode_profile.title

    def seekable(self):
        return False

"""
Rusty audio module forms
"""

class TranscodeForm(Form):
    transcode_profile_slug = SlugField() # TODO: Make it possible to refer to TranscodeProfile.slug
    track_uuid = UUIDField(
        namespace="x-moar:track")
    scrobble = IntegerField(
        required=False,
        default=0)

"""
Rusty audio module models
"""

class MusicLibrary(Model):
    owner = ForeignKey(User)
    name = StringField()
    url = URLField()

class AudioMetadataManager(Manager):
    PRIMARY_KEY = "uuid"
    INDEXES = ("file",),

    
class AudioMetadata(Model):
    objects = AudioMetadataManager()
    
    """
    AudioMetadata represents ID3 tags and stream properties
    """
    
    MIMETYPES = (
        "audio/mpeg",
        "audio/vorbis",
        "audio/flac"
    )
    
    uuid = UUIDField(
        namespace="x-moar:audio_metadata",
        description="Audio metadata UUID",
        null=False)

    file = ForeignKey(
        File,
        transient=True)

    invalid = BooleanField(
        default=False,
        description="Tags invalid")
        
    title = StringField(
        description="Title")
    artist = StringField(
        description="Artist name")
    
    track_number = IntegerField(
        description="Track Number")
    track_count = IntegerField(
        description="Total tracks in this album")
    disc_number = IntegerField(
        description="Disc number")
    disc_count = IntegerField(
        description="Total discs in this album")
    album = StringField(
        description="Album title")
    album_artist = StringField(
        description="Album artist")
    
    # MusicBrainz IDs retrieved from ID3 tags
    artist_mbid = UUIDField(
        namespace = "x-musicbrainz:artist",
        description="MusicBrainz artist identifer")
    release_mbid = UUIDField(
        namespace = "musicbrainz:release",
        description="MusicBrainz release identifier")
    release_artist_mbid = UUIDField(
        namespace="x-musicbrainz:album_artist",
        description="MusicBrainz album artist identifier")
    recording_mbid = UUIDField(
        namespace = "musicbrainz:recording",
        description="Track MBID")
    
    # Acoustid.org trackid lookup
    acoustid_trackid = UUIDField(
        namespace="x-acoustid:trackid",
        description="Acoustid.org track identifier")
        
    # Audio stream metadata
    chromaprint_fingerprint = StringField(
        description="Chromaprint fingerprint")
    audio_md5 = MD5Field(
        description="Decoded audio stream MD5 checksum")
    
    # Acoustid track identifier
    acoustid = UUIDField(
        namespace = "x-acoustid:trackid",
        description="Acoustid track identifier")
    
    duration = FloatField(
        description="Track duration")
    sample_rate = IntegerField(
        description="Sampling rate")
    bit_depth = IntegerField(
        description = "Bit depth")
    channels = IntegerField(
        description = "Channels")
        
    year = YearField(
        description="Year")
    bitrate = IntegerField(
        description="Bitrate")
    codec = StringField(
        max_length=10,
        description="Codec")
    release_country = ForeignKey(
        Country,
        description="Release country")

    def parse(self):
        tags = {}
        metainfo = {}
        
        path = self.file.get_absolute_path()

        self.invalid = True
            
        try:
            if self.file.extension in ('.mp3'):
                metainfo = mutagen.mp3.MP3(path).info
                tags = mutagen.easyid3.EasyID3(path)
                self.codec = "MP3"
            elif self.file.extension in ('.mp4', '.m4a'):
                tags = mutagen.mp4.MP4(path)
                metainfo = tags.info
                self.codec = "MP4"
            elif self.file.extension in ('.ogg'):
                tags = mutagen.oggvorbis.OggVorbis(path)
                metainfo = tags.info
                self.codec = "Vorbis"
            elif self.file.extension in ('.flac'):
                tags = mutagen.flac.FLAC(path)
                metainfo = tags.info 
                self.codec = "FLAC"
            else:
                raise ValueError("Unknown file extension %s" % self.file.extension)
        except mutagen.id3.ID3BadUnsynchData:
            print "Invalid ID3 header for %s" % path
        except mutagen.id3.ID3NoHeaderError:
            print "No ID3 header for %s" % path
        except mutagen.flac.FLACNoHeaderError:
            print "No FLAC header for %s" % path
        except mutagen.mp3.HeaderNotFoundError:
            print "No header found for %s" % path
        except mutagen.oggvorbis.OggVorbisHeaderError:
            print "Invalid OGG Vorbis header for %s" % path
        except mutagen.mp4.MP4StreamInfoError:
            print "Invalid MP4 stream for %s" % path
        else:
            self.invalid = False
        
        if self.invalid:
            return
                        
        # Fetch stream attributes
        self.sample_rate = getattr(metainfo, "sample_rate", None)
        self.duration = metainfo.length
        self.channels = getattr(metainfo, "channels", None)
        self.bitrate = getattr(metainfo, "bitrate", None)
        self.bit_depth = getattr(metainfo, "bits_per_sample", None)

        # Workaround for calculating bitrate
        if not self.bitrate and self.duration:
            self.bitrate = int(self.file.size / self.duration) / 1000 * 8000

        # Artist
        self.artist, = tags.pop("artist", (None,))
        if not self.artist:
            self.artist, = tags.pop("----:com.apple.iTunes:Artist Credit", (None,))
            
        # Track album
        self.album, = tags.pop("album", (None,))
        if not self.album:
            self.album, = tags.pop("----:com.apple.iTunes:Album Credit", (None,))

        # Album artist
        if not self.album_artist:
            self.album_artist, = tags.pop("----:com.apple.iTunes:Album Artist Credit", (None,))

        # Track title
        self.title, = tags.pop("title", (None,))
        
        # Track year
        self.year, = tags.pop("date", (None,))
        if not self.year:
            self.year, = tags.pop("----:com.apple.iTunes:ORIGINAL YEAR", (None,))
            
        def parse_x_of_y(mixed):
            if not mixed:
                return None, None
            elif "www" in mixed:
                return None, None
            elif mixed == "/":
                return None, None
            elif mixed.startswith("/"):
                return None, mixed[1:]
            elif "/" in mixed:
                return mixed.split("/", 1)
            else:
                return mixed, None

        # Disc numbers
        disc, = tags.pop("discnumber", (None,))
        if disc:
            self.disc_number, self.disc_count = parse_x_of_y(disc)
            
        # MusicBrainz identifiers
        self.track_mbid, = tags.pop("musicbrainz_trackid", (None,))
        self.album_mbid, = tags.pop("musicbrainz_albumid", (None,))
        self.album_artist_mbid, = tags.pop("musicbrainz_albumartistid", (None,))
        self.artist_mbid, = tags.pop("musicbrainz_artistid", (None,))
        if not self.track_mbid:
            self.track_mbid, = tags.pop("----:com.apple.iTunes:MusicBrainz Track Id", (None,))
        if not self.artist_mbid:
            self.artist_mbid, = tags.pop("----:com.apple.iTunes:MusicBrainz Artist Id", (None,))
        if not self.album_artist_mbid:
            self.album_artist_mbid, = tags.pop("----:com.apple.iTunes:MusicBrainz Album Artist Id", (None,))
        if not self.album_mbid:
            self.album_mbid, = tags.pop("----:com.apple.iTunes:MusicBrainz Album Id", (None,))

        # Release country
        country_code, = tags.pop("releasecountry", (None,))
        if not country_code:
            country_code, = tags.pop("----:com.apple.iTunes:MusicBrainz Album Release Country", (None,))
        if country_code and country_code.lower() in ("xe", "xw"):
            country_code = None
        try:
            self.release_country = Country.objects.get(code=country_code.lower()) if country_code else None
        except ValueError: # Invalid country code
            self.release_country = None
        
        # Track number could be in various formats: '1', '1/12', 'B2'
        track_number, = tags.pop("tracknumber", (None,))
        self.track_number, self.track_count = parse_x_of_y(track_number)

        if not self.track_number:
            (self.track_number, self.track_count), = tags.pop("trkn", ((None,None),))
        # Consume unused tags
        tags.pop("Genre", None)
        self.invalid = not (self.artist and self.album and self.title)
        
        return tags
        
    def __unicode__(self):
        return "<AudioMetadata of %s>" % self.file.get_absolute_path()

class ArtistManager(Manager):
    INDEXES = ("uuid",), ("mbid",)
    
    def lookup_details(self, artist):
        if artist.mbid and not artist.sort_name:
            ArtistLookupWorker.push(artist)
            
class ArtistField(URNField):
    NAMESPACES = (
        ("x-moar:artist", "Local namespace artist"),
        ("7digital-US:artist", "7Digital artist"), # Supported locales are US, UK and AU
        ("deezer:artist", "Deezer artists"),
        ("discogs:artist", "Discogs artist"), 
        ("emi_bluenote:artist", "EMI Blue Note artist"),
        ("emi_open_collection:artist", "EMI Open Collection artist"),
        ("facebook:artist", "Facebook artist"),
        ("fma:artist", "Free Music Archive artist"),
        ("emi_artists:artist", "EMI artist"),
        ("jambase:artist", "JamBase artist"),
        ("musicbrainz:artist", "Musicbrainz artists"),
        ("playme:artist", "Playme artists"),
        ("rhapsody-US:artist", "Rhapsody artist"),
        ("rdio-US:artist", "Rdio artist"), # Supported locales are AT, AU, BR, CA, CH, DE, DK, ES, FI, FR, IE, IT, NL NO, NZ, PT, SE, UK and US
        ("seatgeek:artist", "SeatGeek artist"),
        ("seatwave:artist", "Seatwave artist"),
        ("songkick:artist", "Songkick artist"),
        ("songmeaningsg:artist", "SongMeanings artist"),
        ("spotify-WW:artist", "Spotify artist"),
        ("twitter:artist", "Twitter artist"),
        ("whosampled:artist", "WhoSampled artist")
    )
    
class TrackField(URNField):
    NAMESPACES = (
        ("x-moar:track", "Local namespace track"),
        ("7digital-US:track", "7Digital track"), # Supported locales are US, UK and AU
        ("deezer:track", "Deezer track"),
        ("emi_bluenote:track", "EMI Blue Note track"),
        ("emi_open_collection:track", "EMI Open Collection track"),
        ("emi_artists:track", "EMI track"),
        ("fma:track", "Free Music Archive track"),
        ("lyricfind-US:song", "LyricFind US song"),
        ("musixmatch-WW:song", "MusixMatch song"),
        ("rhapsody-US:track", "Rhapsody track"),
        ("rdio-US:track", "Rdio track"), # Supported locals are AT, AU, BR, CA, CH, DE, DK, ES, FI, FR, IE, IT, NL NO, NZ, PT, SE, UK and US
        ("playme:track", "Playme track"),
        ("songmeanings:song", "SongMeanings song"),
        ("spotify-WW:track", "Spotify track"),
        ("whosampled:track", "WhoSampled track")
    )
        
class Artist(Model):
    objects = ArtistManager()
    
    uuid = UUIDField(
        namespace="x-moar:artist",
        description="Local artist UUID",
        null=False) # TODO: Use URNField, so it could point to
                    # different namespaces eg.: 
                    # x-moar:artist:slugified-artist-name 
                    # musicbrainz:artist:123412341234-1234-1234-1234-12341234
    mbid = UUIDField(
        namespace="x-musicbrainz:artist",
        description="Artist MusicBrainz identifier")
    name = AudioMetadata.artist
    sort_name = StringField(
        description="Artist sorting key")
    
    annotation = StringField(
        description="Artist description")
    blog = URLField()
    homepage = URLField()
    wikipedia = URLField()
    soundcloud = URLField()
    myspace = URLField()
    country = ForeignKey(Country)
    tags = TagsField()
    
    def __unicode__(self):
        if self.name:
            return "<Artist %s>" % self.name.encode('unicode-escape')
        else:
            return "<Unknown artist>"
            
class AlbumManager(Manager):
    INDEXES = ("uuid",), ("mbid",)
        
class Album(Model):
    objects = AlbumManager()
    
    uuid = UUIDField(
        namespace="x-moar:album",
        description="Local album UUID",
        null=False)
    mbid = UUIDField(
        namespace="x-musicbrainz:release",
        description="MusicBrainz release identifier")
    artist = ForeignKey(
        Artist,
        description="Artist foreign key")
    title = AudioMetadata.album
    track_count = AudioMetadata.track_count
    disc_count = AudioMetadata.disc_count
    country = ForeignKey(
        Country,
        description="Album release country")
    art = URLField()
    hits = IntegerField(
        description="Approximate track binding count",
        default=0,
        transient=True)

    def __unicode__(self):
        if self.title:
            return "<Album %s>" % self.title.encode('unicode-escape')
        else:
            return "<Unknown album>"

class TrackManager(Manager):
    INDEXES = ("uuid",), ("mbid",)

class Track(Model):
    objects = TrackManager()

    file = FileField()
    
    library = ForeignKey(MusicLibrary)

    uuid = UUIDField(
        namespace = "x-moar:track",
        description="Local track UUID",
        null=False)
    mbid = UUIDField(
        namespace = "x-musicbrainz:recording",
        description="MusicBrainz recording identifier")
    artist = ForeignKey(
        Artist,
        description="Artist foreign key")
    album = ForeignKey(
        Album,
        description="Album foreign key")
        
    duration = AudioMetadata.duration
    track_number = AudioMetadata.track_number
    disc_number = AudioMetadata.disc_number
    title = AudioMetadata.title
        
    def __unicode__(self):
        return "<Track %s by %s>" % (
            self.title.encode("unicode-escape") if self.title else "Untitled track",
            self.artist)
    
class TranscodeProfile(Model):
    uuid = None
    slug = SlugField(
        namespace = "x-moar:transcode_profile")
    options = StringField()
    title = StringField()
    mimetype = MimetypeField(
        default="application/octet-stream")
        
    def __unicode__(self):
        return u"<TranscodeProfile %s>" % self.title
        
class TrackBindingManager(Manager):
    INDEXES = ("track", "audio_md5"),
    
    def __init__(self, **properties):
        super(TrackBindingManager, self).__init__(**properties)
        self.cache = {}
        
    def append(self, instance):
        super(TrackBindingManager, self).append(instance)
        i = int(instance.audio_md5, 16)
        try:
            self.cache[i].append(instance)
        except KeyError:
            self.cache[i] = [instance]
        return instance

    def filter_by_audio_md5(self, audio_md5):
        return self.cache.get(int(audio_md5, 16), ())
        
    def total_size(self):
        ts = 0
        for instance in self:
            ts += instance.file.size
        return ts

class TrackBinding(Model):
    """
    TrackBinding associates one or many AudioMetadata objects (referring to actual files) 
    with one or many virtual Track objects (referring single and/or a track on a album/collection
    """
    objects = TrackBindingManager()
    
    created=None
    track=ForeignKey(
        Track)
    audio_md5=AudioMetadata.audio_md5
    file = FileField()

"""
Rusty audio module workers
"""

class ArtistLookupWorker(Worker):
    def handle(self, artist):
        response = musicbrainzngs.get_artist_by_id(artist.mbid, includes=[
            "tags", "annotation", "url-rels", "artist-rels"])

        response = response.get("artist")

        country_code = response.get("country", None)
        if country_code and country_code.startswith("X"):
            # Ignore XE = Europe, XW = Worldwide, etc.
            country_code = None

        annotation = response.get("annotation")
        if isinstance(annotation, dict):
            annotation = annotation.get("text")
            
        properties = dict(
            annotation=annotation,
            name=response.get("name"),
            sort_name=response.get("sort-name"),
            tags=[tag.get("name") for tag in response.get("tag-list",())],
            country=country_code
        )

        for url in reversed(response.get("url-relation-list", ())):
            if url.get("type") == "blog":
                properties["blog"] = url.get("target")
            elif url.get("type") == "soundcloud":
                properties["soundcloud"] = url.get("target")
            elif url.get("type") == "official homepage":
                properties["homepage"] = url.get("target")
            elif url.get("type") == "wikipedia":
                properties["wikipedia"] = url.get("target")
            elif url.get("type") == "myspace":
                properties["myspace"] = url.get("target")
        artist.update(**properties)

class AcoustidLookupWorker(Worker):
    def handle(self, metadata):
        print "Looking up fingerprint for:",  metadata.file

        fh = urllib.urlopen(
            "http://api.acoustid.org/v2/lookup",
            "client=KJ3YSMGL&meta=releases+recordings+tracks&duration=%(duration)d&fingerprint=%(chromaprint_fingerprint)s" % metadata)
        buf = fh.read()
        try:
            response = json_decode(buf)
        except ValueError:
            print "Failed to parse Acoustid server response:", dir(fh), buf
            return
        fh.close()
        
        any_matches = False
            
        for result in response.get("results", ()):
            if result["score"] < 0.7:
                continue
            for recording in result.get("recordings", ()):
                # Update Acoustid object
                metadata.update(
                    acoustid_trackid=result["id"],
                    recording_mbid=recording["id"]
                )

                releases = recording.get("releases", ())
                
                if not releases:
                    print Colors.RED + "!!! AcoustidLookupWorker: No releases for", metadata.file, Colors.END
                for release in releases:
                    for artist in release["artists"]:
                       break

                    # Fetch album artist
                    try:
                        album_artist = Artist.objects.get(
                            mbid=artist["id"])
                    except Artist.DoesNotExist:
                        print "Creating artist %s (%s)" % (artist["name"], artist["id"])
                        album_artist = Artist.objects.create(
                            mbid=artist["id"],
                            name=artist["name"])

                    # Iterate over discs
                    if not "mediums" in release:
                        print Colors.RED + "!!! AcoustidLookupWorker: No mediums for", metadata.file, Colors.END
                    for medium in release["mediums"]:
                        # Iterate over tracks
                        for track in medium["tracks"]:
                            # Iterate over artists
                            for artist in track["artists"]:
                            
                                try:
                                    artist = Artist.objects.get(
                                        mbid=artist["id"])
                                except Artist.DoesNotExist:
                                    print "Creating artist %s (%s)" % (artist["name"], artist["id"])
                                    artist = Artist.objects.create(
                                        mbid=artist["id"],
                                        name=artist["name"])
                                        
                                try:
                                    album = Album.objects.get(
                                        mbid=release["id"])
                                except Album.DoesNotExist:
                                    print "Creating album %s (%s)" % (release["title"], release["id"])
                                    album = Album.objects.create(
                                        mbid=release["id"],
                                        artist=album_artist,
                                        title=release["title"],
                                        track_count=medium["track_count"],
                                        disc_count=release["medium_count"],
                                        hits=0)

                                try:
                                    track = Track.objects.get(
                                        mbid=recording["id"])
                                except Track.DoesNotExist:
                                    print "Creating track:", track["title"]
                                    track = Track.objects.create(
                                        mbid=recording["id"],
                                        artist=artist,
                                        album=album,
                                        duration=metadata.duration,
                                        title=track["title"],
                                        track_number=track["position"],
                                        disc_number=medium["position"]
                                    )

                                # Associate this physical audio file with virtual track object
                                track_binding = TrackBinding.objects.get_or_create(
                                    track=track,
                                    audio_md5=metadata.audio_md5)
                                    
                                track_binding.update(
                                    file=metadata.file)
                                    
                                album.update(hits=album.hits+1)
                                    
                                any_matches = True
            break
            
        if not any_matches:
            if metadata.invalid:
                print Colors.RED + "AcoutsidLookupWorker: Won't use invalid IDv3 metadata as fallback", Colors.END
            else:
                print Colors.RED + "Using IDv3 tags as fallback for:",metadata.file, Colors.END
            
            artist = Artist.objects.get_or_create(
                name=metadata.artist,
                mbid=None)
                
            try:
                album = Album.objects.get(
                    title=metadata.album,
                    artist=artist,
                    mbid=None)
            except Album.DoesNotExist:
                print "Creating album %s (no MusicBrainz association)" % metadata.album
                album = Album.objects.create(
                    title=metadata.album,
                    artist=artist,
                    track_count=metadata.track_count,
                    disc_count=metadata.disc_count,
                    mbid=None,
                    hits=1)

            track = Track.objects.get_or_create(
                title=metadata.title,
                artist=artist,
                album=album,
                duration=metadata.duration,
                track_number=metadata.track_number,
                disc_number=metadata.disc_number,
                mbid=None)

            # Associate this physical audio file with virtual track object
            TrackBinding.objects.get_or_create(
                track=track,
                audio_md5=metadata.audio_md5)
                    
            album.update(hits=album.hits+1)

class AudioMetadataWorker(Worker):
    def handle(self, metadata):
        dirty = False
        
        # Check if we got the tags from directory index or we need to parse this file
        if not metadata.title:
            try:
                metadata.parse()
                dirty = True
            except ValueError:
                print "!!! Failed to parse", metadata.file

        if not metadata.audio_md5 or not metadata.chromaprint_fingerprint:
            path = metadata.file.get_absolute_path()
            
            checksum = hashlib.md5()
            fper = chromaprint.Fingerprinter()
            with audioread.audio_open(path) as stream:
                endposition = stream.samplerate * stream.channels * int(stream.duration)
                try:
                    fper.start(stream.samplerate, stream.channels)
                except chromaprint.FingerprintError:
                    print "!!! Failed to fingerprint", metadata.file, "duration was", stream.duration
                    fper = None
                position = 0
                for chunk in stream:
                    checksum.update(chunk)
                    if position >= endposition:
                        continue
                    elif fper:
                        fper.feed(chunk)
                    position += len(chunk)
                    
                metadata.update(
                    audio_md5=checksum,
                    duration=stream.duration,
                    sample_rate=stream.samplerate,
                    channels=stream.channels,
                    chromaprint_fingerprint=fper.finish() if fper else None)
                    
                if metadata.chromaprint_fingerprint and len(metadata.chromaprint_fingerprint) < 10:
                    print "!!! Warning maybe too short fingerprint:", metadata.chromaprint_fingerprint
                dirty = True
                    

        # Save directory index if we picked up any new files
        if dirty:
            DirectoryIndex.objects.get(directory=metadata.file.directory).save(delay=10)
                    
        # If we have audio MD5
        if metadata.audio_md5:
            track_bindings = TrackBinding.objects.filter_by_audio_md5(metadata.audio_md5)
            # TODO: Upgrade reference, eg. mp3 -> flac
            if track_bindings:
                for track_binding in track_bindings:
                    track_binding.update(
                        file=metadata.file)
                    track_binding.track.album.update(
                        hits=track_binding.track.album.hits+1)
                return

                
        # If this file is not associated with any track, try to lookup
        AcoustidLookupWorker.push(metadata)
            
class DirectoryIndexWorker(Worker):
    """
    Handle directory indexing in a separate thread
    """
    def handle(self, root):
        for directory in root.recurse().filter(lambda o:o.is_directory):
            print "Indexing:", directory
            index = DirectoryIndex.objects.update_or_create(directory=directory)
            for item in index.directory.list():
                if item.mimetype and item.mimetype.startswith("image/"):
                    if not index.cover or "front" in item.filename.lower():
                        index.cover = item

                # Iterate over audio files
                if item.mimetype and item.mimetype in AudioMetadata.MIMETYPES:

                    # Fetch metadata for this file or create one if needed
                    metadata = AudioMetadata.objects.get_or_create(
                        file=item)

                    # If we have audio stream MD5, try to associate this audio
                    # file with tracks in our database
                    if metadata.audio_md5:

                        track_bindings = TrackBinding.objects.filter_by_audio_md5(metadata.audio_md5)                    
                        # TODO: Upgrade reference, eg. mp3 -> flac

                        if track_bindings:

                            for track_binding in track_bindings:
                                track_binding.update(
                                    file=item)
                                track_binding.track.album.update(
                                    hits=track_binding.track.album.hits+1)
                            continue

                    AudioMetadataWorker.push(metadata)

"""
Rusty audio module views
"""

class MusicLibraryView(View):
    def list(self, request):
        return MusicLibrary.objects. \
            filter(lambda library:not library.owner or request.user == library.owner)

class TranscodeProfileView(View):
    def list(self, request):
        return TranscodeProfile.objects. \
            filter(lambda profile:profile.mimetype.startswith("audio/"))                

class AudioMetadataView(View):
    def list(self, request):
        return AudioMetadata.objects.all()

class TrackView(View):
    def list(self, request):
        return Track.objects.all()
        
    def retrieve(self, request, track_uuids):
    

        parser = ForeignKey(Track)
        def iterator_factory():
            for uuid in track_uuids.split(","):
                yield parser.parse(uuid)
        
        return QuerySet(iterator_factory)

    def scrobble(self, request, track_uuid=None, library_uuid=None, artist=None, album=None, title=None):
        library = MusicLibrary.objects.get_or_none(uuid=library_uuid)
        track = Track.objects.get_or_none(uuid=track_uuid)
        
        try:
            lastfm_token = LastFmToken.objects.get(user=request.user)
            lastfm_token.scrobble(title, album, artist)
        except LastFmToken.DoesNotExist:
            pass
        
        return {"message":"ok"}

    def search(self, request, query, uuids_only=False):
        """
        Track search API call
        """
        queryset = Track.objects. \
            filter_by_query(query). \
            limit(500)
        if uuids_only:
            return queryset.map(lambda track:track.uuid)
        else:
            return queryset.map(lambda track: dict(
                    library=track.library.uuid if track.library else None,
                    uuid=track.uuid,
                    title=track.title,
                    artist=track.artist.name if track.artist else None,
                    album=track.album.title if track.album else None,
                    codec=track.codec,
                    bitrate=track.bitrate,
                    sample_rate=track.sample_rate,
                    length=track.length,
                    number=track.track_number,
                    year=track.year))
                
    @TranscodeForm.validate
    def transcode(self, request, transcode_profile_slug, track_uuid):
        """
        Transcode and stream one track identified by the UUID of the track
        """
        try:
            transcode_profile = TranscodeProfile.objects.get(
                slug=transcode_profile_slug)
        except TranscodeProfile.DoesNotExist:
            return Response(status=500, body="No such transcode profile: %s" % transcode_profile_slug)

        try:
            track = Track.objects.get(uuid=track_uuid)
        except Track.DoesNotExist:
            return Response(status=500, body="No such track: %s" % track_uuid)

        try:
            bindings = TrackBinding.objects.filter(lambda binding:binding.track==track).all()
        except TrackBinding.DoesNotExist:
            return Response(status=500, body="Could not find any track bindings")
        
        binding = None
        for binding in bindings:
            try:
                binding.file
                break
            except TrackBinding.DoesNotExist:
                print "Found obsolete binding", binding.track
                binding = None
                
        if not binding:
            return Response(status=500, body="No suitable track binding available (yet), perhaps the indexer is still busy?")
        return Response(
            body=TranscodeStream(binding.file.get_read_stream(), transcode_profile),
            headers={"Content-Type": transcode_profile.mimetype})

class ArtistView(View):
    """
    Artist listing API call
    """
    def retrieve(self, request):
        return Artist.objects. \
            filter(lambda artist: artist.get_albums()). \
            map(lambda artist: dict(
                uuid=artist.uuid,
                name=artist.name,
                albums=artist.get_album_count()))
                
    def list(self, request):
        return Artist.objects.all()
        
    def albums(self, request, artist_uuid):

        artist = Artist.objects.get(
            uuid=artist_uuid)
        
        # Search by album artist
        albums = set(Album.objects. \
            filter(lambda album:album.artist == artist and album.hits > 3))
            
        # Include also "appears on" albums
        for track in Track.objects.filter(lambda track:track.artist == artist):
            albums.add(track.album)

        return [Album._fields["uuid"].serialize(album.uuid, namespace=True) for album in albums]
        
class AlbumView(View):
    """
    Album listing API call
    """
    def list(self, request):
        return Album.objects.all()

    def retrieve(self, request, album_uuids):
        parser = ForeignKey(Album)
        def iterator_factory():
            for uuid in album_uuids.split(","):
                yield parser.parse(uuid)
        return QuerySet(iterator_factory)

    def tracks(self, request, album_uuid):
        album = Album.objects.get(uuid=album_uuid)
        return Track.objects. \
            filter(lambda track:track.album == album). \
            map(lambda track:Track._fields["uuid"].serialize(track.uuid, namespace=True))

    def art(self, request, album_uuid, size=None):
        if size:
            size = int(size)
            assert 32 <= size <= 256, "Invalid thumbnail size %dx%d" % size

        album = Album.objects.get(uuid=album_uuid)
        
        if image:
            if size:
                return image.get_thumbnail(size).download()
            else:
                return image.download()
        else:
            return NotFoundResponse()

    def download(self, request, album_uuid):
        album = Album.objects.get(uuid=album_uuid)
        artist = album.get_artist()
        filename = "%s - %s" % \
            (artist.name if artist else "Various Artists", album.title)
        transcode_profile = TranscodeProfile.objects.get(slug="mp3-vbr")
        
        pipeline = album.get_tracks() | \
            Map(lambda track:track.metadata.file) | \
            Tar() | \
            Gzip(5)
        return Response(pipeline, headers={
                "Content-Type": "application/x-gtar",
                "Content-Disposition": "attachment; filename=\"%s.tar.gz\"" % filename})
        

# Bind fields
MusicLibrary.prepare()
Album.prepare()
Artist.prepare()
Track.prepare()
TrackBinding.prepare()
TranscodeProfile.prepare()
AudioMetadata.prepare()

# If artist is created (eg. by mbid), fetch it's bio & stuff
Artist.objects.created.connect(
    Artist.objects.lookup_details)

# Create transcode profiles
TranscodeProfile.objects.create(
    slug="ogg-vbr", title="OGG Vorbis (VBR)", mimetype="audio/ogg",
    options="-acodec libvorbis -aq 2 -f ogg")
TranscodeProfile.objects.create(
    slug="mp3-vbr", title="MP3 (VBR)", mimetype="audio/mpeg",
    options="-acodec libmp3lame -aq 2 -f mp3")            
TranscodeProfile.objects.create(
    slug="mp3-128k", title="MP3 (128k)", mimetype="audio/mpeg",
    options="-acodec libmp3lame -ab 128k -f mp3")    

