
import time
import re
import os
from md5 import md5
from uuid import uuid4
from rusty.jason import json_decode
from datetime import datetime, timedelta
from rusty.forms import Form
from rusty.views import View
from rusty.http import FoundResponse, Response, ForbiddenResponse
from rusty.decorators import serialize
from rusty.signals import Signal
from rusty.application import Application
from rusty.fields import StringField, DateTimeField, EmailField, ForeignKey, \
    PathField, IntegerField
from rusty.models import Model, Manager

class FacebookLoginForm(Form):
    state = StringField(
        required=False)
    code = StringField(
        required=False)

class LastFmLoginForm(Form):
    token = StringField(
        required=False)

class UserManager(Manager):
    PRIMARY_KEY = "email"
    INDEXES = ("email",),
                
class User(Model):
    objects = UserManager()
    
    email = EmailField()
    first_name = StringField(
        max_length=30)
    last_name = StringField(
        max_length=30)

    def __repr__(self):
        return "<User %s>" % self.email

class FacebookTokenManager(Manager):
    INDEXES = ("user",), ("user","application")

class FacebookToken(Model):
    objects = FacebookTokenManager()
    
    user = ForeignKey(
        User)
    application = ForeignKey(
        Application)
        
    id = IntegerField(
        description="Facebook user id")
    email = EmailField(
        description="Facebook user e-mail")
    username = StringField(
        description="Facebook username if user has created one")
    access_token = StringField(
        description="Access token")
    expires = DateTimeField()
            
    def __repr__(self):
        return "<Facebook access token %s for %s>" % (self.access_token, self.email)
        
class UnixAccountManager(Manager):
    PRIMARY_KEY = "uid"
    INDEXES = ("username",),
    
    def sync(self, extra_uids=()):
        import pwd        
        for p in pwd.getpwall():
            if p.pw_uid not in extra_uids:
                if p.pw_uid < 1000: # Skip system users
                    continue
                if p.pw_uid > 60000:
                    continue
                
            full_name = None
            email = None

            if p.pw_gecos:
                if "," in p.pw_gecos:
                    try:
                        full_name, _, _, _, email = p.pw_gecos.split(",")
                    except ValueError, e:
                        full_name, _ = p.pw_gecos.split(",", 1)

            self.get_or_create(username = p.pw_name).update(
                uid = p.pw_uid,
                gid = p.pw_uid,
                full_name = full_name,
                email = email,
                home = p.pw_dir,
                shell = p.pw_shell)

class UnixAccount(Model):
    objects = UnixAccountManager()
    
    uid = IntegerField()
    gid = IntegerField()
    
    user = ForeignKey(User)
    email = EmailField(
        description="UNIX user e-mail address")
    username = StringField(
        description="UNIX user name")
    home = PathField(
        description="UNIX user home directory")
    full_name = StringField()
    shell = StringField()
    
    def __repr__(self):
        return "<UnixAccount %s%s>" % (
            self.username,
            " (%s)" % self.full_name if self.full_name else "")
            
    def serialize(self):
        email= "lauri.vosandi@gmail.com"
        return dict(
            avatar = "http://www.gravatar.com/avatar/%s" % md5(email.strip().lower()).hexdigest(),
            uid = self.uid,
            gid = self.gid,
            email = self.email,
            username=self.username,
            full_name=self.full_name)
    
    def authenticate(self, challenge):
        import pam
        if pam.authenticate(self.username, challenge, "sshd"):
            token = UnixToken.objects.get_or_create(
                unix_account=self)
            token.update(expires = datetime.now() + timedelta(days=30))
            return token
        else:
            raise UnixToken.DoesNotExist
            
    def switch(self):
        if os.getuid() != 0:
            raise OSError("Only root can switch user")

        # Remove group privileges
        os.setgroups([])

        # Try setting the new uid/gid
        os.setgid(self.gid)
        os.setuid(self.uid)

        # Ensure a very conservative umask
        old_umask = os.umask(077)

class UnixTokenManager(Manager):
#    PRIMARY_KEY = "local_account"
    
    def __init__(self):
        self.group = None
        super(UnixTokenManager, self).__init__()
        
    def set_group(self, group):
        assert not self.group
        self.group = group
        import grp
        for username in grp.getgrnam(group).gr_mem:
            pass

class UnixToken(Model):
    objects = UnixTokenManager()
    
    user = ForeignKey(User)
    unix_account = ForeignKey(UnixAccount)
    expires = DateTimeField(
        description="UNIX user token expiration date and time")
        
class LastFmTokenManager(Manager):
    PRIMARY_KEY = "username"
    INDEXES = ("user",),

class LastFmToken(Model):
    objects = LastFmTokenManager()
    
    application = ForeignKey(
        Application,
        description="Application to which this token was issued")
    user = ForeignKey(
        User)
    username = StringField(
        description="Last.fm user name")
    session_key=StringField(
        description="Last.fm API session key")
    expires=DateTimeField(
        description="Last.fm API session key expiration time")
        
    def update_now_playing(self, track=None, album=None, artist=None):
        """
        Issue updateNowPlaying API call using this token
        """
        self.call("track.updateNowPlaying",
            timestamp=str(int(time.time())),
            track=track,
            artist=artist,
            album=album)

    def scrobble(self, track=None, album=None, artist=None):
        """
        Issue scrobble API call using this token
        """
        self.call("track.scrobble",
            timestamp=str(int(time.time())),
            track=track,
            artist=artist,
            album=album)
    
    def call(self, method, **arguments):
        from urllib2 import urlopen
        from urllib import urlencode
        arguments["method"] = method
        arguments["format"] = "json"
        arguments["sk"] = self.session_key
        arguments["api_key"] = request.application.lastfm_api_key

        url="http://ws.audioscrobbler.com/2.0/"
        buf=""
        for key, value in sorted(arguments.items(), key=lambda (key,value):key):
            if key == "format": # wtf?
                continue
            buf += key
            buf += value.encode("utf-8")
        
        buf += request.application.lastfm_api_secret
        arguments["api_sig"] = md5(buf).hexdigest()
        
        # Come friggin on guys, why don't you support utf-8 in urlencode?!
        fh = urlopen(url, data=urlencode([(key.encode("utf-8"), value.encode("utf-8")) for key,value in arguments.iteritems()]))
        body=fh.read()
        return body 
    
    def __repr__(self):
        return "<Last.fm session key %s for %s>" % (self.session_key, self.username)


class TokenView(View):
    def list(self, request):
        if not hasattr(request, "user"):
            return ForbiddenResponse("Session middleware not enabled")
        if not request.user:
            return ForbiddenResponse("User not authenticated")
        d = dict(
            user=dict(
                uuid=request.user.uuid,
                email=request.user.email,
                first_name=request.user.first_name,
                last_name=request.user.last_name))
        
        # Try to append facebook token   
        try:
            facebook_token = FacebookToken.objects.get(
                user=request.user,
                application=request.application)
            d["facebook"]= dict(
                id=facebook_token.id,
                expires=facebook_token.expires,
                username=facebook_token.username,
                email=facebook_token.email)
        except FacebookToken.DoesNotExist:
            pass
         
        # Try to append Last.fm token   
        try:
            lastfm_token = LastFmToken.objects.get(
                user=request.user,
                application=request.application)
            d["lastfm"]= dict(
                expires=lastfm_token.expires,
                username=lastfm_token.username)
        except LastFmToken.DoesNotExist:
            pass

        # Try to append UNIX account token
        try:
            token = UnixToken.objects.get(
                user=request.user)
            d["unix"] = dict(
                username=token.unix_account.username,
                expires=token.expires)
        except UnixToken.DoesNotExist:
            pass
        return d
        
class UnixAccountView(View):
    def list(self, request):
        return UnixAccount.objects.filter(lambda account:account.email)
        
    def login(self, request, username=None, password=None):
        account = UnixAccount.objects.get(username=username);
        token = account.authenticate(password)
        if token:
            # The account is associated with an user
            if token.user:
                # If session's user is different, enforce user change for session
                if request.user != token.user:
                    request.session.update(
                        user=token.user)
            # The account has no user associated with it
            else:
                # Take over session's user
                if request.user:
                    token.update(user=request.user)
                    
                # Otherwise create user for UNIX account and session
                else:

                    first_name, last_name = account.full_name.split(" ", 1)
                    user = User.objects.get_or_create(
                        first_name=first_name,
                        last_name=last_name,
                        email=account.email)
                        
                    token.update(user=user)
                    request.session.update(user=user)
                
            return dict(
                result=True,
                uid = account.uid,
                gid = account.gid,
                email = account.email,
                token_expires=token.expires,
                home=account.home,
                shell=account.shell,
                full_name=account.full_name)
        else:
            return dict(result=False)


class FacebookView(View):
    def __init__(self, secret=None, callback_url=None, success_url=None):
        self.secret = secret
        self.success_url = success_url
        self.callback_url = callback_url
                
    @FacebookLoginForm.validate
    def login(self, request, state=None, code=None):
        app_id = request.application.facebook_id
        app_secret = self.secret
        redirect_uri = self.callback_url

        if request.user:
            token = FacebookToken.objects.get_or_none(
                user=request.user,
                application=request.application)
        else:
            token = None
        
        if not token:

            state = state or uuid4()

            if not code:
                print "We don't have the code, redirect to facebook"
                return FoundResponse(
                    "https://www.facebook.com/dialog/oauth?scope=email&client_id=%(app_id)s&redirect_uri=%(redirect_uri)s&state=%(state)s" % locals())
            else:
                print "We got code, retrieving access token"
                url = "https://graph.facebook.com/oauth/access_token?client_id=%(app_id)s&redirect_uri=%(redirect_uri)s&client_secret=%(app_secret)s&code=%(code)s" % locals()
                
                from urllib import urlopen
                body = urlopen(url).read()
                
                m = re.match("access_token=(?P<access_token>[0-9a-zA-Z]+)&expires=(?P<expires>\d+)", body)
                try:
                    access_token, expires = m.groups()
                except Exception, e:
                    print "Failed to parse as facebook token:", body
                    raise e
                expires = datetime.now() + timedelta(seconds=int(expires))

                
                fh = urlopen("https://graph.facebook.com/me?access_token=%(access_token)s" % locals())
                buf = fh.read()
                me = json_decode(buf)
                
                token = FacebookToken.objects.get_or_create(
                    email=me["email"],
                    application = request.application)
                token.update(
                    id=me["id"],
                    email=me["email"],
                    username=me.get("username", None),
                    access_token=access_token,
                    expires=expires)

                if not token.user:
                    try:
                        token.update(user=User.objects.get(
                            email=me["email"]))
                    except User.DoesNotExist:
                        token.update(user=User.objects.create(
                            email=me["email"],
                            first_name=me["first_name"],
                            last_name=me["last_name"]))

        if request.user != token.user:
            request.session.update(
                user=token.user)
            request.user = token.user

        return FoundResponse(self.success_url)
            
class LastFmView(View):
    def __init__(self, callback_url=None):
        self.callback_url = callback_url

    @LastFmLoginForm.validate
    def login(self, request, token=None):
        if token:
            # TODO: Deduplicate code here
            args = dict(
                api_key=request.application.lastfm_api_key,
                format="json",
                method="auth.getSession",
                token=token)
                
            def lastfm_hack(args):
                from urllib import urlopen, urlencode
                url="http://ws.audioscrobbler.com/2.0/"
                buf=""
                for key in sorted(args.keys()):
                    if key == "format":
                        continue
                    buf += key
                    buf += args[key]
                
                buf += request.application.lastfm_api_secret
                args["api_sig"]=md5(buf).hexdigest()
                fh = urlopen(url, data=urlencode(args))
                body=fh.read()
                return body
                

            
            body= lastfm_hack(args)
            body = json_decode(body)
            
            username=body["session"]["name"]
            session_key=body["session"]["key"]
            
            lastfm_token = LastFmToken.objects.get_or_create(
                application=request.application,
                username=username)
            lastfm_token.update(
                session_key=session_key,
                user=request.user)
                   
            return Response("%s" % lastfm_token)
            
        
        return FoundResponse("http://www.last.fm/api/auth/?api_key=%s&cb=%s" % (
            request.application.lastfm_api_key, self.callback_url))
            
User.prepare()
UnixAccount.prepare()
UnixToken.prepare()
FacebookToken.prepare()
LastFmToken.prepare()
