import re
import types
import hashlib
from rusty.http import Response, MethodNotAllowedResponse
from rusty.pipeline import Element

class InvalidRoute(Exception):
    pass

class Controller(object):
    def __init__(self, *mixed):
        self.controllers = []
        self.add_controllers(*mixed)

    def __repr__(self):
        return unicode(self).encode("unicode_escape")

    def __unicode__(self):
        return "<Controller>"
        
    def route(self, **properties):
        if self.controllers:
            for mixed in self.controllers:
                if not mixed:
                    continue
                if hasattr(mixed, "route"): # This object knows how to route further
                    try:
                        routed, properties = mixed.route(**properties)
                        return (self,) + routed, properties
                    except InvalidRoute:
                        pass
                elif hasattr(mixed, "dispatch"): # Nope, this object knows how to handle a request
                    return (self, mixed), properties
                else:
                    raise Exception("Duh, this should not happen! %s has no route() nor dispatch()" % mixed)
        raise InvalidRoute()
        
    def dump_chain(self, indent=0):
        print indent*"|   " + "+-", repr(self)

        for child in self.controllers:
            if isinstance(child, Controller):
                child.dump_chain(indent=indent+1)                
            else:
                print indent*"|   " + "    +-", repr(child)
            if hasattr(child, "_get_methods"):
                for name, func in child._get_methods():
                    if name in ("schema", "dispatch"):
                        continue
                    print indent*"|   " + "        def", name

                
    def add_controllers(self, *mixed):
        for m in mixed:
            if isinstance(m, types.GeneratorType):
                self.add_controllers(*m)
            elif isinstance(m, Controller):
                self.controllers.append(m)
            else:
                self.controllers.append(m)

class RegexMappedController(Controller):
    """
    Regex mapped controller uses regular expression to match arbitrary data of a HTTP request
    """
    def __init__(self, pattern, *controllers):
        super(RegexMappedController, self).__init__(*controllers)
        self.pattern = pattern

    def route(self, **propertis):
        raise NotImplementedError()
        
    def __unicode__(self):
        return "<RegexMappedController %s>" % self.pattern

class VirtualHostController(RegexMappedController):
    """
    Virtual host controller uses regular expression to match the host name of the HTTP request
    """
    def route(self, hostname="localhost", **properties):
        m = re.match(self.pattern, hostname)
        if m:
            return Controller.route(self, **properties)
        else:
            raise InvalidRoute()
        
    def __unicode__(self):
        return "<VirtualHostController %s>" % self.pattern

class PathController(RegexMappedController):
    """
    Path controller uses regular expression to match the path of the HTTP request
    """
    def route(self, **properties):
        path = properties.get("path", "")
        m = re.match(self.pattern, path)
        if m:
            properties["path"] = path[m.end():]
            properties.update(m.groupdict())
            return Controller.route(self, **properties)
        else:
            raise InvalidRoute()

    
    def __unicode__(self):
        return "<PathController %s>" % self.pattern

class WebSocketController(object):
    def dispatch(self, request, *args, **kwargs):
        from base64 import b64decode, b64encode
        upgrade = request.headers.get("Upgrade", None).strip()
        assert upgrade == "websocket", "Won't accept upgrade '%s'" % upgrade
        if request.method != "GET":
            return MethodNotAllowedResponse()
            
        websocket_key = request.headers["Sec-WebSocket-Key"]
        websocket_accept = websocket_key + "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
        websocket_accept = hashlib.sha1(websocket_accept).digest()
        websocket_accept = b64encode(websocket_accept)
        
        request.connection.user = request.user
        request.connection.must_upgrade = True
        
        
        return Response(status=101, headers={
            "Upgrade": "websocket",
            "Connection": "Upgrade",
            "Sec-WebSocket-Accept": websocket_accept,
#            "Sec-WebSocket-Protocol": "chat",
#            "Sec-WebSocket-Version": 13
        })

