
import re
import hashlib
from rusty.utils import fqcn
from uuid import UUID, uuid4
from datetime import datetime, date
    
class Field(object):
    def __init__(self, default=None, required=False, null=True, lazy=False, description=None, transient=False):
        self.transient = transient      # Should this field be stored
        self.default = default          # The default value for this field
        self.required = required        # Is this field required
        self.description = description  # Description of this field
        self.null = null                # May the value of this field be None
        self.lazy = lazy                # Whether validators should run while assigning field a value or when the value is being read
        
    def prefill(self):
        if self.default != None:
            return self.default
        elif self.null:
            return None
        else:
            raise ValueError("No default value provided for %s, nor null allowed" % self)
        
    def validate(self, mixed):
        if mixed == None and self.null:
            return None
        else:
            return self.parse(mixed)

    def __repr__(self):
        return "<%s>" % (self.__class__.__name__)
        
    def schema(self):
        d = {"class":fqcn(self)}
        if self.description:
            d["description"] = self.description
        if self.default != None:
            d["default"] = self.default
        if self.required:
            d["required"] = True
        return d
        
    def serialize(self, value, **properties):
        """
        Convert field to any of the primitives (str, int, float, bool)
        """
        return value
                    
class ForeignKey(Field):
    def __init__(self, model, lazy=False, **properties):
        super(ForeignKey, self).__init__(lazy=lazy, **properties)
        self.model = model
        
    def parse(self, mixed):
        from rusty.models import Model
        if isinstance(mixed, Model):
            return mixed
        elif isinstance(mixed, basestring) or isinstance(mixed, int):
            try:
                return self.model.objects.get(**{self.model.objects.PRIMARY_KEY:mixed})
            except Model.DoesNotExist:
                raise ValueError("Could not lookup primary key, %s for foreign key to %s" % (mixed, self.model))
        elif not mixed and self.null:
            return None
        raise ValueError("Don't know how to interpret %s as foreign key" % mixed)
        
    def serialize(self, instance, namespace=False):
        if not instance:
            return None
        pk = getattr(instance, self.model.objects.PRIMARY_KEY)
        if not pk:
            return None
        try:
            field = instance._fields[self.model.objects.PRIMARY_KEY]
        except KeyError:
            raise ValueError("No field %s bound for %s" % (self.model.objects.PRIMARY_KEY, instance))
        return field.serialize(pk, namespace=namespace)
        
    def schema(self):
        """
        Return the schema of referred field
        """
        d = self.model._fields[self.model.objects.PRIMARY_KEY].schema()
        d["description"] = self.description
        return d
        
class StringField(Field):
    def __init__(self, min_length=0, max_length=None, **properties):
        super(StringField, self).__init__(**properties)
        self.min_length = min_length
        self.max_length = max_length
        
    def parse(self, mixed):
        if isinstance(mixed, basestring):
            if self.min_length and len(mixed) < self.min_length:
                raise ValueError("%s shorter than %d characters" % (mixed, self.min_length))
            if self.max_length and len(mixed) > self.max_length:
                raise ValueError("%s longer than %d characters" % (mixed, self.max_length))
            return mixed
        raise ValueError("Don't know how to interpret %s (%s) as string" % (mixed, type(mixed)))
    
    def schema(self):
        d = Field.schema(self)
        d["type"] = "string"
        return d
        
class SlugField(StringField):
    pass
        
class PathField(StringField):
    pass
        
class RegexField(StringField):
    def __init__(self, pattern=None, **properties):
        super(RegexField, self).__init__(**properties)
        if pattern:
            self.pattern = re.compile(pattern)
        
    def parse(self, mixed):
        if isinstance(mixed, basestring):
            if self.pattern.match(mixed):
                return mixed
        raise ValueError("Don't know how to validate %s (%s) with %s regex" % (mixed, type(mixed), self.pattern.pattern))
        
    def schema(self):
        d = StringField.schema(self)
        d["pattern"] = self.pattern.pattern
        return d
        
class IPv4Field(RegexField):
    pattern = re.compile("((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$")
    
class NetmaskField(Field):
    """
    TODO: Create Netmask class to store netmask internally as an int?
    """
    NETMASKS = (
        ( 8,  "255.0.0.0"),
        ( 9,  "255.128.0.0"),
        (10,  "255.192.0.0"),
        (11,  "255.224.0.0"),
        (12,  "255.240.0.0"),
        (13,  "255.248.0.0"),
        (14,  "255.252.0.0"),
        (15,  "255.254.0.0"),
        (16,  "255.255.0.0"),
        (17,  "255.255.128.0"),
        (18,  "255.255.192.0"),
        (19,  "255.255.224.0"),
        (20,  "255.255.240.0"),
        (21,  "255.255.248.0"),
        (22,  "255.255.252.0"),
        (23,  "255.255.254.0"),
        (24,  "255.255.255.0"),
        (25,  "255.255.255.128"),
        (26,  "255.255.255.192"),
        (27,  "255.255.255.224"),
        (28,  "255.255.255.240"),
        (29,  "255.255.255.248"),
        (30,  "255.255.255.252"),
        (31,  "255.255.255.254"),
        (32,  "255.255.255.255"))
            
    def parse(self, mixed):
        if isinstance(mixed, basestring):
            for cidr, netmask in self.NETMASKS:
                if mixed == netmask:
                    return netmask
        elif isinstance(mixed, int):
            for cidr, netmask in self.NETMASKS:
                if mixed == cidr:
                    return netmask
        raise ValueError("Invalid IPv4 netmask %s" % mixed)
        
    def serialize(self, value, **properties):
        for cidr, netmask in self.NETMASKS:
            if value == netmask:
                return cidr
        raise ValueError("Invalid IPv4 netmask %s" % value)

class MimetypeField(RegexField):
    pattern = re.compile("[-\w\.]+/[-\w\.]+$")
        
class EmailField(RegexField):
    pattern = re.compile(
        r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*"
        r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-\011\013\014\016-\177])*"'
        r')@((?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)$)'
        r'|\[(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\]$', re.IGNORECASE)
        
    def schema(self):
        d = StringField.schema(self)
        d["format"] = "email"
        return d

class DomainField(RegexField):
    pattern = re.compile('[a-z0-9]+([\-\.]{1}[a-z0-9]+)*\.[a-z]{2,5}$')
    
    def schema(self):
        d = StringField.schema(self)
        d["format"] = "domain"
        return d

class URLField(RegexField):
    pattern = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|'  # ...or ipv4
        r'\[?[A-F0-9]*:[A-F0-9:]+\]?)'  # ...or ipv6
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)
        
    def schema(self):
        d = StringField.schema(self)
        d["format"] = "uri"
        return d
        
class SlugField(StringField):
    def __init__(self, namespace=None, **properties):
        super(SlugField, self).__init__(**properties)
        self.namespace = namespace
        
    def parse(self, mixed):
        if isinstance(mixed, basestring):
            if self.namespace:
                if mixed.startswith("urn:"):
                    namespace, mixed = mixed.rsplit(":", 1)
                    if namespace != "urn:" + self.namespace:
                        raise ValueError("Invalid namespace:", namespace, "expected:", self.namespace) 
            return mixed
        raise ValueError("Don't know how to interpret %s as slug" % mixed)
        
    def serialize(self, value, namespace=False):
        if not value:
            return None
        elif self.namespace and namespace:
            return "urn:" + self.namespace + ":" + value
        else:
            return value

        
class UUIDField(StringField):
    def __init__(self, namespace=None, **properties):
        super(UUIDField, self).__init__(**properties)
        self.namespace = namespace

    def parse(self, mixed):
        if isinstance(mixed, UUID):
            return mixed
        elif isinstance(mixed, basestring):
            if self.namespace:
                if mixed.startswith("urn:"):
                    namespace, mixed = mixed.rsplit(":", 1)
                    if namespace != "urn:" + self.namespace:
                        raise ValueError("Invalid namespace:", namespace, "expected:", self.namespace)
            if re.match("[0-9a-f]{8}\-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", mixed):
                return UUID(mixed)
            elif re.match("[0-9a-f]{32}", mixed):
                return UUID(hex=mixed)
        raise ValueError("Don't know how to interpret %s (%s) as UUID" % (mixed, type(mixed)))
        
    def prefill(self):
        if self.default:
            return self.default
        elif self.null:
            return None
        else:
            return uuid4()
            
    def schema(self):
        d = StringField.schema(self)
        d["format"] = "uuid"
        return d
        
    def serialize(self, uuid, namespace=False):
        if not uuid:
            return None
        elif self.namespace and namespace:
            return "urn:" + self.namespace + ":" + str(uuid)
        else:
            return str(uuid)
            

class MD5Field(StringField):
    """
    TODO: Optimize and use integer internally
    """
    def parse(self, mixed):
        if isinstance(mixed, long):
            return "%032x" % mixed
#            return mixed
        if isinstance(mixed, basestring):
            if re.match("[0-9a-f]{32}", mixed):
                return  mixed
#                return int(mixed, 16)
        if hasattr(mixed, "hexdigest") and callable(mixed.hexdigest):
            return mixed.hexdigest()
#            return int(mixed.hexdigest(), 16)
        raise ValueError("Don't know how to interpret %s (%s) as MD5 checksum" % (mixed, type(mixed)))
        
    def schema(self):
        d = StringField.schema(self)
        d["format"] = "md5"
        return d

    def serialize(self, value, **properties):
        return value
#        return "%032x" % value

class URNField(StringField):
    NAMESPACES = (
        ("uuid", "UUID"),
    )
    
    def parse(self, mixed):
        if isinstance(mixed, basestring):
            for namespace, title in self.NAMESPACES:
                if mixed.startswith("urn:" + namespace):
                    return namespace, mixed[5+len(namespace):]
            namespace, value = mixed.split(":", 1)
            raise ValueError("Namespace %s not found in %s" % namespace, 
                [ns for (ns,_) in self.NAMESPACES])
        raise ValueError("Don't know how to interpret %s as URNField value" % mixed)

    def serialize(self, (namespace, value)):
        return "urn:" + namespace + ":" + value

class DateTimeField(StringField):
    def __init__(self, auto_now=False, **properties):
        super(DateTimeField, self).__init__(**properties)
        self.auto_now = auto_now
        
    def parse(self, mixed):
        if isinstance(mixed, datetime):
            return mixed
        if isinstance(mixed, basestring):
            if re.match("\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+Z$", mixed):
                return datetime.strptime(mixed, "%Y-%m-%dT%H:%M:%S.%fZ")
            elif re.match("\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$", mixed):
                return datetime.strptime(mixed, "%Y-%m-%d %H:%M:%S")
        raise ValueError("Don't know how to interpret %s as datetime" % mixed)
        
    def prefill(self):
        if self.default != None:
            return self.default
        elif self.auto_now:
            return datetime.now()
        elif self.null:
            return None
        else:
            raise ValueError()
        
    def validate(self, mixed):
        if mixed == None and self.null:
            return None
        return self.parse(mixed)
        
    def serialize(self, value, **properties):
        return value.strftime('%Y-%m-%dT%H:%M:%S.%fZ') if value else None
        
    def schema(self):
        d = StringField.schema(self)
        d["format"] = "date-time"
        return d
        
class IntegerField(Field):
    def __init__(self, minimum=None, maximum=None, **properties):
        self.minimum = minimum
        self.maximum = maximum
        super(IntegerField, self).__init__(**properties)
        
    def parse(self,mixed):
        if isinstance(mixed, int) or isinstance(mixed, long):
            return mixed
        elif isinstance(mixed, basestring):
            if re.match("\d+$", mixed):
                return int(mixed)
        raise ValueError("Don't know how to interpret %s (%s) as integer for %s" % (mixed, type(mixed), self.description))
        
    def schema(self):
        d = Field.schema(self)
        d["type"] = "integer"
        if self.minimum != None:
            d["minimum"] = self.minimum
        if self.maximum != None:
            d["maximum"] = self.maximum
        return d

class PortField(IntegerField):
    pass
    
class FileDescriptorField(IntegerField):
    pass

class FloatField(Field):
    def parse(self, mixed):
        if isinstance(mixed, float):
            return mixed
        elif isinstance(mixed, int):
            return float(mixed)
        elif isinstance(mixed, basestring):
            return float(mixed)
        raise ValueError("Don't know how to interpret %s as float" % mixed)
        
    def schema(self):
        d = Field.schema(self)
        d["type"] = "float"
        return d
        
class BooleanField(Field):
    def parse(self, mixed):
        if isinstance(mixed, bool):
            return mixed
        elif isinstance(mixed, basestring):
            if mixed.lower() in ("1", "true", "yes"):
                return True
            elif mixed.lower() in ("0", "false", "no"):
                return False
        raise ValueError("Don't know how to interpret %s as boolean" % mixed)
        
    def schema(self):
        d = Field.schema(self)
        d["type"] = "boolean"
        return d

class TagsField(Field):
    def parse(self, haystack):
        if isinstance(haystack, basestring):
            haystack = haystack.split("|") # Split by pipe symbol
        if isinstance(haystack, list) or isinstance(haystack, tuple):
            haystack = [needle.lower().strip() for needle in haystack] # Convert to lowercase and strip spaces
            return [needle for needle in haystack if needle] # Skip empty tags
        else:
            raise ValueError("Don't know how to interpret %s (%s) as list of items")
            
    def serialize(self, tags, **properties):
        return "|".join(tags) if tags else None

class ListField(Field):
    def __init__(self, model, **properties):
        super(ListField, self).__init__(**properties)
        self.model = model
        self.fk = ForeignKey(model)
        
    def parse(self, haystack):
        from rusty.models import Model
        if isinstance(haystack, basestring):
            haystack = haystack.split(",")
        if isinstance(haystack, list) or isinstance(haystack, tuple):
            instances = ()
            for needle in haystack:
                if isinstance(needle, Model):
                    instances = instances + (needle,)
                elif isinstance(needle, basestring):
                    instances = instances + (self.model.objects.get(**{self.model.objects.PRIMARY_KEY:needle}),)
                else:
                    raise ValueError("Don't know how to interpret %s as list field item" % needle)
            return instances
        elif not haystack and self.null:
            return None
        else:
            raise ValueError("Don't know how to interpret %s (%s) as list of items" % (haystack, type(haystack)))
            
    def serialize(self, instances, namespace=False, **properties):
        return [self.fk.serialize(instance, namespace=namespace) for instance in instances]
        
class CoordinateField(Field):
    def parse(self, mixed):
        if isinstance(mixed, basestring):
            mixed = [float(j) for j in mixed.split(",")]
        if isinstance(mixed, list) or isinstance(mixed, tuple):
            latitude, longitude = mixed
        return latitude, longitude
        
class JSONField(Field):
    def parse(self, mixed): # Make sure we contain only primitives?
        return mixed

class YearField(Field):
    def parse(self, mixed):
        if not mixed:
            return None
        elif isinstance(mixed, int):
            if mixed >= 1000 and mixed < 3000:
                return mixed
        elif isinstance(mixed, basestring):
            if re.match("[123][0-9][0-9][0-9]$", mixed):
                return int(mixed)
            elif re.match("[123][0-9][0-9][0-9]\-[0-9][0-9]\-[0-9][0-9]$", mixed):
                return int(mixed[:4])
            elif re.match("[0-9][0-9]\-[0-9][0-9]\-[123][0-9][0-9][0-9]$", mixed):
                return int(mixed[-4:])
            elif re.match("[123][0-9][0-9][0-9] [0-9][0-9] [0-9][0-9]$", mixed):
                return int(mixed[:4])
            elif re.match("[123][0-9][0-9][0-9]/[123][0-9][0-9][0-9]", mixed): # dafuq?!
                return None
            elif re.match("\d{4}", mixed): #dafuq?!
                return None
        raise ValueError("Don't know how to interpret %s (%s) as year" % (mixed, type(mixed)))

class DateField(StringField):
    def __init__(self, auto_today=False, **properties):
        super(DateField, self).__init__(**properties)
        self.auto_today = auto_today
        
    def parse(self, mixed):
        if isinstance(mixed, date):
            return mixed
        if isinstance(mixed, basestring):
            if re.match("\d{4}-\d{2}-\d{2}$", mixed):
                mixed = datetime.strptime(mixed, "%Y-%m-%d")
        if isinstance(mixed, datetime):
            return mixed.date()
        raise ValueError("Don't know how to interpret %s as date" % mixed)
        
    def prefill(self):
        if self.default != None:
            return self.default
        elif self.auto_today:
            return date.today()
        elif self.null:
            return None
        else:
            raise ValueError()
        
    def validate(self, mixed):
        if mixed == None and self.null:
            return None
        return self.parse(mixed)
        
    def serialize(self, value, **properties):
        return value.strftime('%Y-%m-%d') if value else None
        
    def schema(self):
        d = StringField.schema(self)
        d["format"] = "date"
        return d

