

import os
import unicodedata
from rusty.files import File
from rusty.streams import InputFileStream
from rusty.http import Response, NotFoundResponse, ForbiddenResponse, \
    FoundResponse
from rusty.templating import JinjaResponse
from rusty.controllers import Controller, InvalidRoute
from rusty.views import View

class JailedController(View):
    def __init__(self, jail="static/", template=None):
        self.jail = File.objects.get_by_path(jail)

class DirectoryIndexView(JailedController):
    template = "directory_index.html"
    
    def route(self, **properties):
        path = properties.get("path", "")
        if path and not path.endswith("/"):
            raise InvalidRoute("No slash suffix")
        absolute_path = os.path.realpath(os.path.join(self.jail.get_absolute_path(), path))
        if not os.path.isdir(absolute_path):
            raise InvalidRoute("Resolved path is not directory")
        return (self,), properties
        
    def list(self, request):
        for file_object in request.directory.list():
            if file_object.filename.startswith("."):
                continue
            if file_object.filename.endswith("~"):
                continue
            if file_object.is_directory:
                record = dict(
                    type="directory",
                    filename=file_object.filename,
                    owner=file_object.owner)
            elif file_object.is_file:
                record = dict(
                    type="file",
                    filename=file_object.filename,
                    extension=file_object.extension,
                    mimetype=file_object.mimetype,
                    size=file_object.size,
                    owner=file_object.owner)
            else:
                continue
            yield record
        
    def dispatch(self, request):
        absolute_path = os.path.realpath(os.path.join(self.jail.get_absolute_path(), request.path))
        request.directory = File.objects.get_by_path(absolute_path)
        assert request.directory.parent, "I ain't got no jail"
        request.directory.invalidate()
        request.path = ""
        
        if request.method == "GET" and not request.headers.has_key("X-Requested-With"):
            return self.render(request)
        return super(DirectoryIndexView, self).dispatch(request)
        
    def render(self, request):

        if not request.directory.is_directory:
            return NotFoundResponse("File %s not found" % absolute_path)

        stack = ()
        current = request.directory

        while current.inode != self.jail.inode:
            print "Comparing:", current, "with", self.jail
            if not current:
                break
            stack = (current,) + stack
            current = current.parent

        directories = sorted(request.directory.list_directories(), key=lambda f:f.filename)
        files = sorted(request.directory.list_files(), key=lambda f:f.filename)
        items = directories + files
        context = {
            "stack": [(d.filename, (len(stack)-i-1)*"../") for i,d in enumerate(stack)],
            "directory": request.directory,
            "items": items
        }
        return JinjaResponse(template=self.template, context=context)
        
    def __repr__(self):
        return "<DirectoryIndexView at %s>" % self.jail.get_absolute_path()

class FileServeView(JailedController):
    def route(self, **properties):
        path = properties.get("path", "")
        if not path:
            raise InvalidRoute()
        absolute_path = os.path.realpath(os.path.join(self.jail.get_absolute_path(), path))
        if not absolute_path.startswith(self.jail.get_absolute_path()):
            raise InvalidRoute()
        if not os.path.isfile(absolute_path):
            raise InvalidRoute()
        return (self,), properties

    def resolve_path(self, request):
        absolute_path = os.path.realpath(os.path.join(self.jail.get_absolute_path(), request.path))
        if not absolute_path.startswith(self.jail.get_absolute_path()):
            raise File.DoesNotExist
        if not os.path.isfile(absolute_path):
            raise File.DoesNotExist
            
        return File.objects.get_by_path(absolute_path)
        
    def dispatch(self, request):
        file_object = self.resolve_path(request)

        assert file_object, "Invalid file: %s" % file_object
        return Response(file_object) #, headers={"Cache-Control": "max-age=3600"})
        
    def __repr__(self):
        return "<FileServeView at %s>" % self.jail.get_absolute_path()

