
import os
import stat
import mimetypes
import unicodedata
from threading import Timer
from rusty.jason import json_decode, json_encode
from uuid import uuid4
from datetime import datetime
from rusty.signals import Signal
from rusty.http import Response
from rusty.authentication import UnixAccount
from rusty.fields import DateTimeField, ForeignKey, StringField, IntegerField
from rusty.models import Model, Manager, QuerySet
from rusty.streams import InputStream
from rusty.volume import Volume
from PIL import Image, ExifTags

"""
Hardcore bloatware for caching os.stat() calls
"""

class FileManager(Manager):
    primary_key = None

    def __init__(self):
        super(FileManager, self).__init__()
        self.inode_to_file_mapping = {}

    def get_by_path(self, path):
        if isinstance(path, str):
            path = path.decode('utf-8')
            
        try:
            s = os.stat(path)
        except OSError, e:
            print e.errno
            if e.errno == 2:
                raise File.DoesNotExist(u"File %s does not exist" % path)
            else:
                raise e
        
        pair = (s.st_dev, s.st_ino)
        
        if pair in self.inode_to_file_mapping:
            return self.inode_to_file_mapping[pair]
        else:
            path = os.path.realpath(path)
            common = dict(
                volume = Volume.objects.get_by_device(s.st_dev),
                filename = os.path.basename(path),
                modified = datetime.fromtimestamp(s.st_mtime),
                changed = datetime.fromtimestamp(s.st_ctime),
                inode = s.st_ino)

            parent_path = os.path.dirname(os.path.realpath(path))
            parent = self.get_by_path(parent_path) if parent_path != u"/" else None
            

            if stat.S_ISREG(s.st_mode):
                mimetype, _ = mimetypes.guess_type(path)
                _, extension = os.path.splitext(path)

                retval = File(
                    directory = parent,
                    extension = extension.lower(),
                    mimetype = mimetype,
                    size = s.st_size,
                    **common)
            elif stat.S_ISDIR(s.st_mode):
                retval = Directory(
                    parent = parent,
                    **common)
            elif stat.S_ISSOCK(s.st_mode):
                retval = UnixDomainSocket(
                    directory = parent,
                    **common)
            else:
                raise Exception("Unknown file type for %s" % path)
                
            self.inode_to_file_mapping[pair] = retval
            self.objects.append(retval)
            return retval

class AbstractFile(Model):
    """
    Abstract file, the common class for file and directory
    """
    is_hidden = property(lambda f: f.filename.startswith(".") or f.filename.endswith("~"))
    is_file = False
    is_directory = False
    is_socket = False
    
    # Disable UUID and created timestamp for this model
    created = None
    uuid = None
    
    volume = ForeignKey(Volume)
    filename = StringField()
    modified = DateTimeField()
    changed = DateTimeField()
    inode = IntegerField(
        transient=True)
    size = IntegerField()
    owner = ForeignKey(UnixAccount)

    def __hash__(self):
        return self.volume.device << 32 | self.inode
    
    @property
    def icon(self):
        if self.is_directory:
            return "places/22/folder.svg"
        elif self.mimetype:
            return "mimes/22/" + self.mimetype.replace('/', '-') + ".svg"
        else:
            return "mimes/22/unknown.svg"

class File(AbstractFile):
    """
    inotify friendly object representation of a file
    """
    
    is_file = True
    
    objects = FileManager()
    
    def __init__(self, directory=None, extension=None, mimetype=None, size=0, exists=True, **kwargs):
        AbstractFile.__init__(self, **kwargs)
        self.size = size
        self.directory = directory
        self.extension = extension
        self.mimetype = mimetype
        self.exists = exists

    def get_thumbnail(self, size=64):
        assert self.mimetype.startswith("image/"), "Unable to generate thumbnail from %s" % self.get_absolute_path()
        cache = self.directory.get_or_create_directory(u".thumbnail")
        reduced = cache.get_or_create_file(u"%s-%dx%d.png" % (self.filename, size, size))
        if not reduced.exists:
            i = Image.open(self.get_read_handle())
            
            exif=dict((ExifTags.TAGS[k], v) for k, v in i._getexif().items() if k in ExifTags.TAGS)
                
            orientation = exif.get('Orientation', 1)
            if orientation == 6:
                i=i.rotate(270, expand=True)
            elif orientation == 3:
                i=i.rotate(180, expand=True)
            elif orientation == 8:
                i=i.rotate(90, expand=True)
            
            i.thumbnail((size, size), Image.ANTIALIAS)
            i.save(reduced.get_write_handle(), format="png")
            print u"Generating %s ..." % reduced.get_absolute_path()
        return reduced
        
    def download(self, inline=False):
        """
        Return HTTP response containing this file as body
        """
        props = {"Content-Type": self.mimetype}
        if self.filename and not inline:
            props["Content-Disposition"] = "attachment; filename=\"%s\"" % \
                unicodedata.normalize('NFKD', self.filename).encode('ascii', 'ignore')
        return Response(
            body=self,
            headers=props)
        
    def get_read_handle(self):
        return open(self.get_absolute_path(), "rb")
        
    def get_read_stream(self):
        return InputStream(self.get_read_handle(), filename=self.filename, mimetype=self.mimetype)
        
    def get_write_handle(self, overwrite=False):
        return open(self.get_absolute_path(), "w")
        
    def get_absolute_path(self):
        return os.path.join(
            self.directory.get_absolute_path() if self.directory else
            self.volume.mountpoint, self.filename)
            
    def __repr__(self):
        return "<File %s>" % self.get_absolute_path().encode('unicode-escape')
        
class FileField(ForeignKey):
    def __init__(self):
        super(FileField, self).__init__(File, transient=True)
    
class UnixDomainSocket(AbstractFile):
    is_socket = True
    
    def __init__(self, directory=None, **kwargs):
        AbstractFile.__init__(self, **kwargs)
        self.directory = directory
            
class Directory(AbstractFile):
    is_directory = True
    mimetype = None
    size = None
    
    def __init__(self, parent=None, **kwargs):
        AbstractFile.__init__(self, **kwargs)
        self.parent = parent
        self._children = None
        
    def get_or_create_file(self, filename):
        try:
            return self.get(filename)
        except File.DoesNotExist:
            return File(directory=self, filename=filename, exists=False)
        
    def get_or_create_directory(self, filename):
        try:
            return self.get(filename)
        except File.DoesNotExist:
            os.makedirs(os.path.join(self.get_absolute_path(), filename))
            self._children = None
            return Directory(parent=self, filename=filename)
                
    def exists(self, filename):
        for child in self.list():
            if child.filename == filename:
                return True
        return False    
        
    def __repr__(self):
        return "<Directory %s>" % self.get_absolute_path().encode('unicode-escape')

    def get_absolute_path(self):
        return os.path.join(self.parent.get_absolute_path() if self.parent else "/", self.filename) 

    def _recurse(self, func):
        if not func(self):
            return
        else:
            yield self
        for child in self.list():
            if child.is_directory:
                if func(child):
                    for obj in child._recurse(func):
                        yield obj
            else:
                yield child

    def recurse(self, func=lambda o:not o.is_hidden):
        """
        Return generator which recurses across the directory tree and yields File or Directory object for each entry
        """
        return QuerySet(lambda: self._recurse(func))
        
    def invalidate(self):
        self._children = None
        
    def list_files(self):
        return self.list(lambda o:o.is_file and not o.is_hidden)
        
    def list_directories(self):
        return self.list(lambda o:o.is_directory and not o.is_hidden)

    def list(self, func=None):
        if not self._children:
            p = self.get_absolute_path()
            self._children = [File.objects.get_by_path(os.path.join(p,c)) for c in os.listdir(p)]
        for child in self._children:
            if func and not func(child):
                continue
            yield child


class DirectoryIndexManager(Manager):
    PRIMARY_KEY = "directory"
    INDEXES = ("directory",),
    
    bindings = ()
    
    def update_or_create(self, directory):
        index = self.get_or_create(directory=directory)
        index.update()
        return index
            
    def bind(self, key, model):
        """
        Bind model to directory index props
        """
        self.bindings += (key, model),
        
    def update_related(self, key, model, instance):
        print "Updated related:", instance
            
class DirectoryIndex(Model):
    objects = DirectoryIndexManager()
    
    comment = StringField()
    directory = ForeignKey(File)
    cover = FileField() # Image file for this directory
    
    path = property(lambda self:os.path.join(self.directory.get_absolute_path(), ".props.json"))
    
    def update(self):
        items = self.directory.list()
        all_props = self.read_props()
        if os.path.exists(self.path):
            for item in items:
                item_props = all_props.get("files",{}).get(item.filename, None)
                if item_props:
                    for key, model in self.objects.bindings:
                        if key in item_props:
                            props = item_props[key]
                            props["file"] = item
                            instance = model.objects.create(**props)
            
    def read_props(self):
        props = {}
        try:
            fh = open(self.path)
            props = json_decode(fh.read())
            fh.close()
        except IOError:
            pass # File didn't exist
        except ValueError:
            pass # Failed to parse JSON
            
        if "files" not in props:
            props["files"] = {}

        return props
        
    def save(self, delay=0):
        """
        Serialize and store file model and related model fields
        """
        # Try to cancel timer if we have one
        try:
            self.timer.cancel()
        except AttributeError:
            pass

        if delay:
            self.timer = Timer(delay, self.save)
            self.timer.start()
            return
            
        props = self.read_props()
        for item in self.directory.list():
            if item.filename not in props["files"]:
                props["files"][item.filename] = {}
                
            props["files"][item.filename]["file"] = item.serialize()
            
            for key, model in self.objects.bindings:
                try:
                    props["files"][item.filename][key] = model.objects.get(file=item).serialize()
                except Model.DoesNotExist:
                    pass
                    
        buf = json_encode(props)
        fh = open(self.path, "w")
        fh.write(buf)
        fh.close()
        print "Wrote directory index of", len(buf), "bytes to", self.directory
        
        self.updated.emit(self)

DirectoryIndex.prepare()
File.prepare()
