
from rusty.fields import Field

class Form(object):
    """
    Hardcore bloatware for validating properties and other stuff
    """
    def __init__(self, **properties):
        # Initialize properties
        self._properties = {}
        self._bind_fields()
        
        # Set of lazy fields whose value has been already validated
        self._validated = set()
        
        # Validate and set property values
        for key, value in properties.iteritems():
            self.set(key, value)
        
        
    def _prefill_fields(self):
        # Set default values
        for key, field in self._fields.iteritems():
            if key not in self._properties:
                if not hasattr(field, "default"):
                    continue
                self._properties[key] = field.prefill()
                    
    @classmethod
    def _bind_field(cls, key, field):
        def getter(self):
            try:
                value = self._properties[key]
            except KeyError:
                mixed = self._fields[key].default
                return mixed() if callable(mixed) else mixed
                
            if field.lazy and key not in self._validated:
                value = field.validate(self._properties[key])
                self._properties[key] = value
                self._validated.add(key)
            return value

        def setter(self, value):
            if not field.lazy:
                value = field.validate(value)
            else:
                self._validated.discard(key)

            self._properties[key] = value
        setattr(cls, key, property(getter, setter))
        
    @classmethod
    def _get_fields(cls):
        return dict([ \
            (name, attribute) \
            for (name, attribute) \
            in [ \
                (key, getattr(cls,key)) \
                for key \
                in dir(cls)] \
            if isinstance(attribute, Field) ])
        
    @classmethod
    def _bind_fields(cls):
        # If we have already bound fields, return
        if hasattr(cls, "_fields"):
            return

        # Store our fields
        cls._fields = cls._get_fields()
            
        for key, field in cls._fields.iteritems():
            cls._bind_field(key, field)

    @classmethod
    def validate(cls, target_func):
        """
        Decorator used to validate view arguments with this particular form
        """
        def decorated(instance, request, **kwargs):
            kwargs.update(request.query or {})
            form = cls(**kwargs)
            return target_func(instance, request, **form.items())
        decorated.form = cls
        return decorated
        
    @classmethod
    def schema(cls):
        """
        Return JSON schema of this form
        """
        cls._bind_fields()
        print "Getting schema:",dict([(name,field.schema()) for name,field in cls._fields.items()])
        return {
            "class":"%s.%s" % (cls.__module__, cls.__name__),
            "properties":dict([(name,field.schema()) for name,field in cls._fields.items()])}
        
    def __getitem__(self, key):
        return self.get(key)
            
    def __iter__(self):
        return self._properties.iteritems()
        
    def set(self, key, value):
        """
        Validate and set field value
        """
        if key in self._fields:
            try:
                setattr(self, key, value)
            except ValueError:
                raise ValueError("Invalid value %s (%s) for %s of %s" % (value, type(value), key, self))

        
    def get(self, key):
        """
        Get field value
        """
        assert key in self._fields
        return getattr(self, key)
        
    def items(self):
        return self._properties

    def serialize(self, *fields):
        return dict([(key, self.get(key)) for key in fields or self._fields.keys()])

