
# coding: utf-8
import os
import re
import types
import socket
import urllib2
from rusty.utils import urldecode
from rusty.jason import json_decode, json_encode, JasonDecodeError
from subprocess import Popen, PIPE
from datetime import datetime, timedelta
from sendfile import sendfile
from splicetee import splice, SPLICE_F_MORE, SPLICE_F_NONBLOCK
from rusty.streams import InputStream
from rusty.terminal import Colors

KERNEL_NAME, nodename, release, version, machine = os.uname()
KERNEL_VERSION, _ = release.split("-", 1)

lsb_release = Popen(["/usr/bin/lsb_release", "-a"], stdout=PIPE).communicate()[0]

DISTRIBUTION_NAME, DISTRIBUTION_VERSION = re.search(
    "Distributor ID:\s*(.+?)\n.*Release:\s*(.+?)\n",
    lsb_release, flags=re.DOTALL).groups()

class Request(object):
    def __init__(self, method="GET", path="/", hostname="localhost", port=80, headers={}, query={}, connection=None, body=None, secure=False, **properties):
        self.created = datetime.utcnow()
        self.method = method
        self.path = path
        self.hostname = hostname        # Server hostname
        self.port = port                # TCP connection port
        self.headers = headers          # HTTP headers
        self.query = query              # GET/POST arguments
        self.secure = secure            # Whether connection is wrapped in SSL
        self.connection = connection    # Underlying TCP connection
        self.body = body
        self.response = None          # The response this request corresponds to

        # Build base URL, eg. for generating links relative to this request
        self.url_base = "%s://%s%s" % (
            "https" if self.secure else "http",
            self.hostname,
            ":%d" % self.port if self.secure and self.port != 443 or not self.secure and self.port != 80 else "")

        self.url = "%s/%s" % (self.url_base, self.path)
        
        if self.headers.has_key("User-Agent"):
            from rusty.user_agent import UserAgent
            self.user_agent = UserAgent.objects.get_or_create(
                fingerprint=self.headers.get("User-Agent"))
            self.user_agent.hit()
        else:
            self.user_agent = None
        
    def __repr__(self):
        return "<Request %s %s>" % (self.method, self.url)

RESPONSES = {
    100: 'Continue', 101: 'Switching Protocols', 200: 'OK', 201: 'Created',
    202: 'Accepted', 203: 'Non-Authoritative Information', 204: 'No Content',
    205: 'Reset Content', 206: 'Partial Content', 300: 'Multiple Choices',
    301: 'Moved Permanently', 302: 'Found', 303: 'See Other',
    304: 'Not Modified', 305: 'Use Proxy', 307: 'Temporary Redirect',
    400: 'Bad Request', 401: 'Unauthorized', 402: 'Payment Required',
    403: 'Forbidden', 404: 'Not Found', 405: 'Method Not Allowed',
    406: 'Not Acceptable', 407: 'Proxy Authentication Required',
    408: 'Request Timeout', 409: 'Conflict', 410: 'Gone',
    411: 'Length Required', 412: 'Precondition Failed',
    413: 'Request Entity Too Large', 414: 'Request-URI Too Long',
    415: 'Unsupported Media Type', 416: 'Requested Range Not Satisfiable',
    417: 'Expectation Failed', 500: 'Internal Server Error',
    501: 'Not Implemented', 502: 'Bad Gateway', 503: 'Service Unavailable',
    504: 'Gateway Timeout', 505: 'HTTP Version Not Supported'
}

class Response(object):
    def __init__(self, body=None, status=200, headers={}):

        self.request = None           # The request this response corresponds to
        self.transfer_method = None
        self.status = status
        self.headers = {
            "Server": "Rusty/0.1 (%s/%s %s/%s)" % (KERNEL_NAME, KERNEL_VERSION, DISTRIBUTION_NAME, DISTRIBUTION_VERSION),
            "Connection": "keep-alive",
        }
        
        from rusty.files import File
  
        if isinstance(body, unicode):
            body = body.encode('utf-8')

        if isinstance(body, basestring):
            self.transfer_method = "copy"
            self.headers.update({
                "Content-Length": len(body)})
                
        # Body is file with determined length, use sendfile to copy file to the
        # socket and also attach content length header
        elif isinstance(body, File):
            self.transfer_method = "sendfile"
            self.headers.update({
                "Content-Type": body.mimetype,
                "Content-Length": body.size })
            body = body.get_read_stream()
            
        # Body is input stream with undetermined length, eg. transcode stream.
        # Transfer it with splice and close connection to notify browser that
        # transferring the content is finished
        elif isinstance(body, InputStream):
            self.transfer_method = "splice"
            self.headers.update({
                "Connection": "close"})
                
        # Body is generator, eg. streamed JSON.
        elif isinstance(body, types.GeneratorType):
            self.transfer_method = "iterator"
            self.headers.update({
                "Transfer-Encoding": "chunked"})
                
        # No body at all, eg. WebSocket
        elif not body:
            pass
        else:
            raise ValueError("Don't know how to transmit body of type %s" % type(body))

        if self.transfer_method:
            self.headers.update({
                "X-Transfer-Method": self.transfer_method})
            
        if body:
            self.headers.update({
                "X-Body-Type":
                    "%s.%s" % (body.__module__, body.__class__.__name__) \
                    if hasattr(body, "__module__") \
                    else body.__class__.__name__
            })

        self.body = body            
        self.headers.update(headers)
        
    def disable_keepalive(self):
        self.headers["Connection"] = "close"
        
    def render_headers(self):
        if self.request:
            td = datetime.utcnow() - self.request.created
            self.headers["X-Served-In"] = "%.2fms" % (
                td.microseconds / 1000.0 +
                td.seconds * 1000 +
                td.days * 86400000)
        buf = "HTTP/1.1 %d %s\r\n" % (self.status, RESPONSES[self.status])
        for key, value in self.headers.iteritems():
            buf += "%s: %s\r\n" % (key, value)
        buf += "\r\n"
        return buf.encode("ascii")
        
class NotFoundResponse(Response):
    def __init__(self, body="Not found"):
        super(NotFoundResponse, self).__init__(body, 404)
        
class ForbiddenResponse(Response):
    def __init__(self, body="Forbidden"):
        super(ForbiddenResponse, self).__init__(body, 403)
        
class FoundResponse(Response):
    def __init__(self, location):
        super(FoundResponse, self).__init__(
            body="Resource moved to %s" % location,
            status=302,
            headers={"Location":location})
            
class MethodNotAllowedResponse(Response):
    def __init__(self):
        super(MethodNotAllowedResponse, self).__init__(
            body="Method not allowed",
            status=405)
        
class HttpRequestParser(object):
    def __init__(self, connection):
        self.buffer = ""
        self.connection = connection
        self.reset()
        
    def reset(self):
        """
        Reset parser state
        """
        self.method = None
        self.version = None
        self.path = None
        self.query = {}
        self.port = 80
        self.body = ""
        self.headers = None
        
    def parsed_body(self):
        # Check if we have parsed headers yet
        if not self.headers:
            return False

        # Check if we have buffered enough body
        content_length = self.headers.get("Content-Length", 0)
        if len(self.buffer) >= content_length:
            self.body = self.buffer[:content_length]
            self.buffer = self.buffer[content_length:]
        else:
            return False

        # Parse POST arguments
        if self.headers.get("Content-Type", None) == "application/x-www-form-urlencoded":
            self.query.update(urldecode(self.body))
        return True
        
    def parsed_headers(self, verbose=True):
        if self.headers:
            return False
            
        if '\n\n' in self.buffer:
            headers, self.buffer = self.buffer.split("\n\n", 1)
        elif '\n\r\n' in self.buffer:
            headers, self.buffer = self.buffer.split("\n\r\n", 1)
        else:
            return False

        headers = headers.replace("\r\n", "\n").replace("\r", "").split("\n")
        
        # Split method and path
        first_line = headers.pop(0)
        self.method, path, self.version = first_line.split(" ", 3)
                
        if not verbose:
            print "%s %s" % (self.method, path)

        # Parse headers
        headers = [header.split(": ", 1) for header in headers]
        self.headers = dict(headers)
    
        # Parse hostname and port
        hostname = self.headers.get("Host")
        if ":" in hostname:
            hostname, port = hostname.split(":", 1)
            port = int(port)
        else:
            port = 80
        self.hostname = hostname.decode("idna")

        # Parse GET arguments
        if "?" in path:
            path, query = path.split("?", 1)
            self.query = urldecode(query)

        # Parse path
        path = urllib2.unquote(path).decode('utf-8')
        self.path = path[1:]

        if verbose:
            print Colors.BOLD + Colors.BRIGHT_CYAN + ">>>" + Colors.END,
            print Colors.BOLD + self.method + Colors.END,
            print Colors.GRAY + "http://%s:%d"  % (self.hostname, port) + Colors.END + Colors.BOLD + path,
            if self.query:
                print self.query,
            print Colors.END
            for key,value in sorted(headers, key=lambda(key,value):key):
                print Colors.BLUE + ">>>" + Colors.END,
                print (Colors.GRAY if key.startswith("X-") else "") + "%s: %s" % (key,value) + Colors.END

        assert path.startswith("/"), "Incorrect path"
        return True
        
    def pop(self, buffer_size=32768):
        """
        Pop chunk of bytes from the connection read handle
        """
        chunk = self.connection.recv(buffer_size)
            
        if chunk:
            self.buffer += chunk
            
            # We have not parsed the HTTP headers yet,
            # but the buffer contains double-newlines now
            if self.parsed_headers():
                pass
                
            # We have parsed the HTTP headers, check if we have received
            # HTTP request body already
            if self.parsed_body():
                self.connection.server.route(
                    connection=self.connection,
                    method=self.method,
                    hostname=self.hostname,
                    port=self.port,
                    path=self.path,
                    headers=self.headers,
                    query=self.query)
                self.reset()
        else:
            # We did get no bytes, assume client disconnected
            self.connection.forget("No bytes received to parse HTTP request")

class HttpResponseRenderer(object):
    def __init__(self, connection):    
        self.connection = connection    
        self.buffer = ""
        self.body = None
        self.chunked = False
        self.close_connection = False
        
    def pop(self, buffer_size=32768):
        # If there is something in the buffer (headers, iterator remainder)
        # try to send it first
        if self.buffer:
            bytes_sent = self.connection.send(self.buffer)
            self.buffer = self.buffer[bytes_sent:]
        # If the buffer is clean, try to fetch another chunk from iterator
        elif self.body:
            # Body is iterator
            if self.transfer_method == "iterator":
                try:
                    assert not self.buffer, "Buffer not empty"
                    
                    # Try to fill the buffer from body iterator
                    while True:
                        self.buffer += self.body.next()
                        
                        # If the chunk is bigger than buffer size, break
                        if len(self.buffer) >= buffer_size:
                            break
                except StopIteration:
                    self.body = None

                assert self.buffer, "Buffer empty"
                
                # If we are transferring content in chunked mode, 
                # prepend byte count and append newline
                if self.chunked:
                    self.buffer = ("%x\r\n" % len(self.buffer)) + self.buffer + "\r\n"
                    
                    # If this was the last chunk, append empty chunk
                    if not self.body:
                        self.buffer += "0\r\n\r\n"

                # Try to drain the buffer to the socket
                bytes_sent = self.connection.send(self.buffer)
            
                # Keep remainder
                self.buffer = self.buffer[bytes_sent:]
                    
            # Source is seekable, therefore we know the length of the 
            elif self.transfer_method == "sendfile":
                bytes_sent = sendfile(
                    self.connection.fd,
                    self.body.fileno(),
                    None,
                    buffer_size)
                
                # Close file handle
                if not bytes_sent:
                    self.body.close()

            # Source is not seekable, therefore we don't know the length
            # and we have to send disconnect after transferring the body
            elif self.transfer_method == "splice":
                try:
                    _, _, bytes_sent = splice(
                        self.body.fileno(),
                        0,
                        self.connection.fd,
                        0,
                        buffer_size,
                        SPLICE_F_MORE)
                        
                    # Close file handle
                    if not bytes_sent:
                        self.body.close()

                except OSError:
                    bytes_sent = 0
            else:
                raise TypeError("Don't know how to transfer body with method %s" % self.transfer_method)

            # We didn't transmit any bytes therefore we're finished with the body                
            if not bytes_sent:
                self.body = None
        else:
            bytes_sent = 0
                
        # We did not send any bytes, assume connection was closed
        if not bytes_sent:
            if self.close_connection:
                self.connection.server.epoll.modify(self.connection.fd, 0)
                try:
                    self.connection.handle.shutdown(socket.SHUT_RDWR)
                except socket.error:
                    pass
            else:
                self.connection.set_incoming()

        if self.connection.must_upgrade:
            self.connection.upgrade_to_websocket()
            
    def respond(self, response, verbose=True):
    
        # If the response body is a generator, eg. JSON chunks then
        # stream Gzipped body if browser supports compression
        if isinstance(response.body, types.GeneratorType):
            print "------------> Compressing body of:", response.request.path
            accept_encoding = response.request.headers.get('Accept-Encoding', '')
            encodings = [encoding.strip().lower() for encoding in accept_encoding.split(",")]
            
            if "gzip" in encodings or "x-gzip" in encodings:
                from rusty.compression import Gzip
                response.body = Gzip().pipe(response.body)
                response.headers["Content-Encoding"] = "gzip"
                response.headers.update({
                    "X-Transfer-Method": "iterator",
                    "Transfer-Encoding": "chunked"})
                            
        # Write response headers to socket
        headers = response.render_headers()
        status, rest = headers.strip().split("\r\n", 1)
        _, status = status.split(" ", 1)

        if response.status >= 400:
            print Colors.BOLD + Colors.BRIGHT_RED + "<<<",
        else:
            print Colors.BOLD + Colors.BRIGHT_GREEN + "<<<",
        print Colors.END + Colors.BOLD + status,
        print Colors.END
        for line in sorted(rest.split("\r\n")):
            print (Colors.RED if response.status >= 400 else Colors.GREEN) + "<<<" + Colors.END,
            print Colors.GRAY + line + Colors.END if line.startswith("X-") else line
        print
        self.buffer += headers
           
        if response.body:
            self.transfer_method = response.headers.get("X-Transfer-Method")
            self.chunked = response.headers.get("Transfer-Encoding", None) == "chunked"
            content_length = response.headers.get("Content-Length", None)
            if self.chunked:
                if isinstance(response.body, InputStream):
                    self.body = response.body
                else:
                    self.body = iter(response.body)
            elif content_length >= 0:
                if isinstance(response.body, unicode):
                    raise Exception("Encode UTF-8 first!")
                elif isinstance(response.body, str):
                    assert len(response.body) == content_length, "Response body length %d does not match Content-Length %d" % (len(response.body), content_length)
                    self.buffer += response.body
                elif isinstance(response.body, InputStream):
                    self.body = response.body
                else:
                    self.body = iter(response.body)
            elif response.headers.get("Connection", "keep-alive") == "close":
                self.body = response.body
            else:
                raise ValueError("Don't know how to transfer %s ... (%s) as HTTP response body" % (
                    repr(response.body)[:100], type(response.body)))
        
        self.close_connection = response.headers["Connection"] == "close"
        # Mark this connection ready for outbound traffic
        self.connection.set_outbound()
        

