
import re
import os
import GeoIP
import codecs
from rusty.fields import StringField, ForeignKey, IntegerField, StringField, \
    CoordinateField
from rusty.models import Model, Manager
from rusty.views import View

class CountryCodeField(StringField):
    def parse(self, mixed):
        code = super(CountryCodeField, self).parse(mixed)
        return code.lower() if code else ""

class ContinentManager(Manager):
    PRIMARY_KEY = "code"
    INDEXES = ("code",),
    
    def sync(self):
        self.create(code="af", name="Africa")
        self.create(code="as", name="Asia")
        self.create(code="eu", name="Europe")
        self.create(code="na", name="North America")
        self.create(code="sa", name="South America")
        self.create(code="oc", name="Oceania")
        self.create(code="an", name="Antartica")

class Continent(Model):
    objects = ContinentManager()
    
    uuid = None
    created = None
    code = StringField()
    name = StringField()
    
    def __unicode__(self):
        return self.name

class CountryManager(Manager):
    PRIMARY_KEY = "code"
    INDEXES = ("code",),
    
    def load(self, path_iso3166_csv="iso3166.csv", path_geoip_dat="/usr/share/GeoIP/GeoIP.dat"):
        self.load_iso3166(path_iso3166_csv)
        self.gi = GeoIP.open(path_geoip_dat, GeoIP.GEOIP_MEMORY_CACHE)
        
        for country in self.all():
            country_code = country.code.upper()
            try:
                continent_code = GeoIP.country_continents[country_code]
            except KeyError:
                continue                
            country.update(continent=continent_code.lower())
        print "Loaded %d country codes" % self.count()
        
    def get_by_ipv4(self, addr):
        """
        Get country by IP address
        """
        try:
            return self.get(code=self.gi.country_code_by_addr(addr))
        except ValueError:
            raise Country.DoesNotExist()
            
    def get_by_hostname(self, name):
        """
        Get country by hostname
        """
        return self.get(code=self.gi.country_code_by_name(name))
    
    def load_iso3166(self, filename):
        # ISO3166 standardized country codes and names
        # http://www.iso.org/iso/home/standards/country_codes/country_names_and_code_elements_txt.htm
        lines = open(filename).read().decode('utf-8').replace('\r', '').split("\n")
        assert lines[0] == "Country Name;ISO 3166-1-alpha-2 code"
        for line in lines[1:]:
            if not line:
                continue

            name, code = line.split(";")
            code = code.lower()
            name = name.title()
            name = re.sub(r' Of$', r' of', name)
            name = re.sub(r' Of ', r' of ', name)
            
            try:
                self.get(code=code)
            except Country.DoesNotExist:
                self.create(
                    code = code,
                    name = name)

class Country(Model):
    objects = CountryManager()
    
    uuid = None                 # Disable UUIDs for this model
    created = None              # Disable created timestamp
    
    code = CountryCodeField(
        description="The ISO 3166-1 alpha-2 code for the country")
    continent = ForeignKey(Continent)
    slug = StringField(
        description="The short, lowercase version of the name")
    name = StringField(
        description="The name of the country")

    def __unicode__(self):
        return self.name
        
class CityManager(Manager):
    def sync(self):
        fh = codecs.open("GeoLiteCity-Location.csv", encoding="iso8859-1")
        assert "MaxMind LLC" in fh.readline()
        assert "locId,country,region,city,postalCode,latitude,longitude,metroCode,areaCode" in fh.readline()
        cities = 0

        while True:
            line = fh.readline()
            if not line:
                break
            m = re.search(r"(?P<location_id>\d+),\"(?P<country>\w\w)\",\"(?P<region_code>\w*?)\",\"(?P<name>.*?)\",\"(.*?)\",(?P<coordinate>-?\d+\.?\d*,-?\d+\.?\d*),(.*?),(.*?)", line)
            if not m:
                print "Failed to parse:", line
                break
            else:
                properties = m.groupdict()
                if not properties["name"]:
                    continue
                if properties["country"] == "AN":
                    continue
                self.create(**properties)
        print "Loaded %d cities" % cities


class City(Model):
    uuid = None
    created = None
    
    objects = CityManager()
    
    location_id = IntegerField()
    country = ForeignKey(Country)
    name = StringField()
    region = StringField()
#    postal_code = StringField()
#    metro_code = StringField()
#    area_code = StringField()
    coordinate = CoordinateField()
 

class CountryView(View):
    def list(self, request):
        return Country.objects.all()

class CityView(View):
    def list(self, request):
        return City.objects.all()
        
Continent.prepare()
Country.prepare()
City.prepare()
Continent.objects.sync()

