
"""
Django-style model and model manager base classes
"""

import re
from uuid import UUID
from datetime import datetime
from rusty.signals import Signal
from rusty.queryset import QuerySetBase, QuerySet
from rusty.pipeline import Pipeline, PipelineMixin
from rusty.forms import Form
from rusty.fields import UUIDField, DateTimeField, Field
from rusty.utils import fqcn

class Model(Form):
    class DoesNotExist(Exception):
        pass

    abstract = True
    
    uuid = UUIDField(
        null=False,
        description="Random Unique Identifier")
    created = DateTimeField(
        auto_now=True,
        description="Created")
    
    def __init__(self, **properties):
        self.updated = Signal()
        super(Model, self).__init__(**properties)
        self._prefill_fields()

    def __eq__(self, other):
        if self != other:
            assert not other or not hasattr(self, "objects") or getattr(self, self.objects.PRIMARY_KEY) != getattr(other, self.objects.PRIMARY_KEY), \
                "Inconsistency detected - the object instances differ but the primary key is the same"
            return False
        return True
        
    def update(self, **properties):
        """
        Update field values and emit updated signal if any of the non-transient
        fields are updated
        """
        dirty = []
        for key, value in properties.items():
            if key not in self._fields:
                raise KeyError("No field '%s' bound to %s" % (key, self))
            if not self._fields[key].transient:
                dirty.append(key)
            setattr(self, key, value)
        if dirty:
            self.updated.emit(self)
        
    def save(self):
        self.updated.emit(self)
        
    def __unicode__(self):
        return self.__class__.__name__
    
    def __repr__(self):
        return unicode(self).encode("unicode_escape")

    @classmethod
    def prepare(cls):
        """
        Bind fields, ensure default manager and mark this class as not abstract
        """
        cls._bind_fields()
        cls._ensure_default_manager()
        cls.abstract = False
        
    @classmethod
    def _get_name(cls):
        return re.sub(r"(?!^)([A-Z]+)", r"_\1", cls.__name__).lower()

    @classmethod
    def _ensure_default_manager(cls):
        """
        Make sure this model has manager attached to it and that the manager
        is aware of the model which it is attached to
        """
        if not hasattr(cls, "objects"):
            print "Binding default model manager for", cls
            cls.objects = Manager()
        else:
            print "Binding", cls.objects, "for", cls
        cls.objects.set_model(cls)

    @classmethod
    def get_descendants(cls):
        for subclass in cls.__subclasses__():
            yield subclass
            for subclass in subclass.get_descendants():
                yield subclass

    def serialize(self, transient=True, namespace=False):
        return dict([(key, self._fields[key].serialize(self.get(key), namespace=namespace)) for key in self._fields.keys() if transient or not self._fields[key].transient])
        
class Manager(QuerySetBase, PipelineMixin):
    PRIMARY_KEY = "uuid"
    INDEXES = ("uuid",),
    instances = []
    
    def __init__(self):
        self.created = Signal()
        self.updated = Signal()
        
        self.model = None
        self.objects = []
        self.lookup = {}
        self.counter = 0
        self.instances.append(self)
        
    def set_model(self, model):
        self.model = model
        
    def create(self, **props):
        assert self.model, "No model attached to %s!" % self
        return self.append(self.model(**props))
        
    def set_storage(self, storage):
        self.storage = storage
        assert self.model, "No model associated with %s" % self.__class__.__name__
        def save_instance(instance):
            serialized = instance.serialize(transient=False)
            key = serialized.pop(self.PRIMARY_KEY) # Pop primary key!
#            key = str(getattr(instance, self.PRIMARY_KEY))
            assert "\\" not in key, "Invalid primary key for multi-file storage: %s" % key
            storage.replace(
                key.replace("/", "\\"),
                serialized)
        
        keys = storage.list()
        print "Loading", len(keys), self.model.__name__, "instances" 
        for key in keys:
            properties = storage.retrieve(key)
            properties[self.PRIMARY_KEY] = key.replace("\\", "/") # Push primary key!
            self.create(**properties)
        
        self.created.connect(save_instance)
        self.updated.connect(save_instance)
        
    def __iter__(self):
        return iter(self.objects)
        
    def on_instance_updated(self, instance):
        self.updated.emit(instance)
        
    def append(self, instance):
        counter = self.counter = self.counter + 1
        self.objects.append(instance)
        self.created.emit(instance)
        
        # Store references for indexes
        for index in self.INDEXES:
            items = {}
            for key in index:
                items[key] = getattr(instance, key)
            self.lookup[tuple(items.items())] = instance
            
        instance.updated.connect(self.on_instance_updated)
        return instance
        
    def all(self):
        return tuple(self.objects)
        
    def count(self):
        return len(self.objects)

    def get(self, **properties):
        assert self.model, "No model associated with %s" % self
        items = tuple([(key, self.model._fields[key].validate(value)) for (key, value) in properties.iteritems()])
        
        # Try to lookup via index
        if tuple(properties.keys()) in self.INDEXES:
            try:
                return self.lookup[items]
            except KeyError:
                raise self.model.DoesNotExist("Instance %s not found" % properties)
                
        # If there is no corresponding index, what the heck - chew through the whole goddamn list
        else:
            print "!!! Iterating over fields:", ",".join(properties.keys()),
            print "on", fqcn(self.model)
            for obj in self.objects:
                matched = True
                for key, value in items:
                    if getattr(obj, key) != value:
                        matched = False
                        break
                if matched:
                    return obj
            raise self.model.DoesNotExist("Instance %s not found" % properties)
        
    def get_or_create(self, **properties):
        try:
            return self.get(**properties)
        except self.model.DoesNotExist:
            return self.create(**properties)
            
    def get_or_none(self, **properties):
        try:
            return self.get(**properties)
        except self.model.DoesNotExist:
            return None
            
    def first(self):
        return iter(self).next()
        
    def update(self, **properties):
        for obj in self:
            obj.update(**properties)

