
import re

class PipelineMixin():
    def __or__(self, element):
        return Pipeline(self, element)
        
    def __gt__(self, element):
        for needle in self | element:
            pass

class Element(PipelineMixin):
    def dispatch(self, needle):
        raise NotImplementedError()

    def pipe(self, haystack):
        for needle in haystack:
            yield self.dispatch(needle)
        
    def __repr__(self):
        return "<Element>"
        
class Filter(Element):
    def __init__(self, func):
        self.func = func
        
    def __repr__(self):
        return "<Filter>"
        
    def pipe(self, haystack):
        for needle in haystack:
            if self.func(needle):
                yield needle
                
    def __repr__(self):
        return "<Filter>"
        
class Sort(Element):
    def __init__(self, func=None, reverse=False):
        self.func = func
        self.reverse = reverse
        
    def __repr__(self):
        return "<Sort>"
        
    def pipe(self, mixed):
        for obj in sorted(mixed, key=self.func, reverse=self.reverse):
            yield obj
            
class Split(Element):
    def __init__(self, separator=" ", strip=True):
        self.separator = separator
        self.strip = strip
        
    def __repr__(self):
        return "<Split>"
        
    def pipe(self, haystack):
        for needle in haystack:
            if needle:
                yield [s.strip() for s in needle.split(self.separator)]
            
class Cast(Element):
    def __init__(self, *types):
        self.types = types

    def __repr__(self):
        return "<Cast>"
        
    def pipe(self, haystack):
        for needle in haystack:
            if needle:
                yield [self.types[i](obj[i]) for i in range(0, len(self.types))]
        
class Map(Element):
    def __init__(self, func):
        self.func = func
        
    def __repr__(self):
        return "<Map>"
    
    def pipe(self, haystack):
        for needle in haystack:
            if needle:
                yield self.func(needle)

class Slice(Element):
    def __init__(self, begin=0, end=None):
        self.begin = begin
        self.end = end

    def pipe(self, haystack):
        j = 0
        for needle in haystack:
            if j >= self.begin: # TODO: Implement slice end code
                yield needle
            j += 1
        
class Substitute(Element):
    def __init__(self, pattern, replacement):
        self.pattern = re.compile(pattern)
        self.replacement = replacement
        
    def pipe(self, haystack):
        for needle in haystack:
            assert isinstance(needle, basestring)
            yield self.pattern.sub(self.replacement, needle)
            
class Capitalize(Element):
    def pipe(self, haystack):
        for needle in haystack:
            yield needle.title()
  
class Pipeline(PipelineMixin):
    def __init__(self, *elements):
        self.elements = elements

    def __or__(self, target):
        if isinstance(target, Pipeline):
            return Pipeline(*(self.elements + target.elements))
        else:
            return Pipeline(*(self.elements + (target,)))

    def dispatch(self, needle):
        def pop(needle, element, *elements):
            if elements:
                return pop(element.dispatch(needle), *elements)
            else:
                return element.dispatch(needle)
        return pop(*self.elements)

    def execute(self):
        def pop(mixed, element, *elements):
            if elements:
                return pop(element.pipe(mixed), *elements)
            else:
                return element.pipe(mixed)
        return pop(*self.elements)
        
    def pop(self):
        return Pipeline(*(self.elements[:-1])), self.elements[-1]
        
    def __iter__(self):
        return self.execute()
                
    def __repr__(self):
        return "<Pipeline: %s>" % " | ".join([p.__class__.__name__ for p in self.elements])

