
import types
from rusty.pipeline import PipelineMixin

class QuerySetBase(object):
    def map(self, func, unique=False):
        if unique:
            encountered = []
            for obj in self:
                mapped = func(obj)
                if mapped not in encountered:
                    encountered.append(mapped)
                    yield mapped
        else:
            for obj in self:
                yield func(obj)
            
    def limit(self, count):
        def iterator_factory():
            countdown = count
            for obj in self:
                countdown -= 1
                if countdown >= 0:
                    yield obj
                else:
                    return
        return QuerySet(iterator_factory)
        
    def filter(self, func):
        def iterator_factory():
            for obj in self:
                if func(obj):
                    yield obj
        return QuerySet(iterator_factory)
        
    def order_by(self, field="created"):
        return QuerySet(sorted(self), key=lambda o:getattr(o, field))
            
class QuerySet(QuerySetBase, PipelineMixin):
    """
    Lazy-evaluating query set class, similar to Django's
    """
    def __init__(self, mixed):
        if isinstance(mixed, types.TupleType):
            self.objects = mixed
        elif isinstance(mixed, types.ListType):
            self.objects = tuple(mixed)
        elif isinstance(mixed, types.GeneratorType):
            self.objects = tuple(mixed) # Postpone this perhaps?
        elif isinstance(mixed, types.FunctionType):
            self.objects = None
            self.iterator_factory = mixed
        else:
            raise ValueError("Do not know how to construct QuerySet from %s" % mixed)
        super(QuerySet, self).__init__()

    def evaluate(self):
        self.objects = tuple(self.iterator_factory())
        self.iterator_factory = None
        return self
        
    def count(self):
        """
        Count objects in this query set
        """
        if self.objects:
            return len(self.objects)
        else:
            counter = 0
            for obj in self:
                counter += 1
            return counter

    def __iter__(self):
        if self.objects:
            iterator = iter(self.objects)
        else:
            iterator = self.iterator_factory()
            assert isinstance(iterator, types.GeneratorType)
        return iterator
            
    def __repr__(self):
        return "<QuerySet of %d objects>" % self.count()

    def all(self):
        if not self.objects:
            self.objects = tuple(self.iterator_factory())
        return self.objects

