
# coding: utf-8
import os
import array
import struct
import urllib2
import traceback
import types
import select
import socket
from time import sleep
from rusty.events import Event
from rusty.models import Model, Manager
from rusty.http import HttpRequestParser, HttpResponseRenderer
from rusty.location import Country
from rusty.fields import IntegerField, PortField, ForeignKey, IPv4Field, \
    FileDescriptorField
from rusty.files import File
from rusty.streams import InputStream
from rusty.signals import Signal
from rusty.http import Request, NotFoundResponse
from rusty.websocket import WebSocketParser, WebSocketRenderer
from rusty.controllers import InvalidRoute
from rusty.utils import urldecode
from rusty.terminal import Colors
        
class ConnectionManager(Manager):
    PRIMARY_KEY = "fd"
    INDEXES = ("fd",),
    
    def create(self, server, handle, address, reader=None, writer=None):
        return self.append(Connection(server, handle, address, reader, writer))
        
class Connection(Model):
    """
    Universal connection class, which should be easily usable by HTTP server,
    HTTP client, WebSocket server, WebSocket client etc.
    """
    objects = ConnectionManager()
    
    fd = FileDescriptorField(
        description="File descriptor of the connection")
    address = IPv4Field(
        description="Client IPv4 address")
    port = PortField(
        description="Client port")
    country = ForeignKey(Country,
        description="Client country, guessed by client IPv4 address")
#    received = IntegerField(
#        default=0,
#        null=False,
#        description="Bytes received on this connection")
#    transmitted = IntegerField(
#        default=0,
#        null=False,
#        description="Bytes transmitted on this connection")
#    session = ForeignKey(Session)
#    user = ForeignKey(User)

    def __init__(self, server, handle, address, reader=None, writer=None):
        address, port = address
        super(Connection, self).__init__(
            fd=handle.fileno(),
            address=address,
            port=port)

        self.must_upgrade = False
        server.connections[self.fd] = self
        handle.setblocking(0)
        self.server = server
        self.received = 0
        self.transmitted = 0
        
        self.receive_dead = False
        self.transmit_dead = False
        
        # session, user, last user agent?
        
        self.handle = handle
        try:
            self.country = Country.objects.get_by_ipv4(self.address)
        except Country.DoesNotExist:
            pass
        self.input = reader(self)
        self.output = writer(self)
        
        self.server.epoll.register(
            self.fd,
            select.EPOLLIN | select.EPOLLHUP | select.EPOLLERR)
            
    def recv(self, buffer_size=32768):
        try:
            buf = self.handle.recv(buffer_size)
        except socket.error:
            print "No bytes received on arbitrary connection"
            buf = ''
        self.received += len(buf)
        return buf
        
    def send(self, buf):
        # TODO: Set to outgoing mode
        sent_bytes = self.handle.send(buf)
        self.transmitted += sent_bytes
        return sent_bytes
        
    def set_incoming(self):
        """
        Set connection to incoming mode
        """
        self.server.epoll.modify(
            self.fd,
            select.EPOLLIN | select.EPOLLHUP | select.EPOLLERR)
        
    def set_outbound(self):
        """
        Set connection to outbound mode
        """
        self.server.epoll.modify(
            self.fd,
            select.EPOLLOUT | select.EPOLLHUP | select.EPOLLERR)
            
    def set_duplex(self):
        """
        Set connection to full duplex mode
        """
        try:
            self.server.epoll.modify(
                self.fd,
                select.EPOLLIN | select.EPOLLOUT | select.EPOLLHUP | select.EPOLLERR)
        except IOError, e: # Bad file descriptor, disconnected perhaps?
            self.forget("Bad file descriptor, disconnected perhaps")
            
    def forget(self, reason=None):
        """
        Acknowledge client initiated TCP disconnect
        """
        print Colors.BOLD + Colors.BRIGHT_YELLOW + "---" + Colors.END,
        print "Connection (%d) from" % self.fd, self.address,
        if self.country:
            print "(%s)" % self.country,
        print "on port", self.port, "shutdown acknowledged,",
        connection = self.server.connections.pop(self.fd, None)        
        print "now polling %d connections" % len(self.server.connections),
        if reason:
            print "reason:", reason,
            

        print

        self.server.epoll.unregister(self.fd)

        try:
            self.handle.close()
        except socket.error, e:
            print "!!! Unable to close connection:", e
            # Transport endpoint is not connected
            pass



    def shutdown(self):
        """
        Initiate server-side TCP disconnect
        """
        self.server.epoll.modify(self.fd, 0)
        try:
            self.handle.close()
        except socket.error:
            # Transport endpoint is not connected
            pass

    def upgrade_to_websocket(self):
        """
        Upgrade to websocket, this is no way neat implementation!
        """
        
        self.handle.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self.input = WebSocketParser(self)
        self.output = WebSocketRenderer(self)
        self.set_incoming()
        Event.objects.create(
            name="device_connected_event",
            user=self.user, # User is assigned after negotiation to websocket
            origin=self.uuid)
        
    def __repr__(self):
        return "<Connection (%d) %s:%d from %s>" % (self.fd, self.address, self.port, self.country or "N/A")
        
    def serialize(self):
        d = super(Connection, self).serialize()
        d.update(
            rx=self.received,
            tx=self.transmitted,
            downlink_alive=not self.receive_dead,
            uplink_alive=not self.transmit_dead,
            reader=self.input.__class__.__name__,
            writer=self.output.__class__.__name__)
        return d
        

class Reactor(object):
    """
    Generic TCP server
    """
    
    def __init__(self, controller, address="0.0.0.0", port=8000, verbose=False):
        self.verbose = verbose
        self.controller = controller

        self.address = address
        self.port = port
        
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        
        pair = self.address, self.port
        
        # TODO: Bind port to certain parser+writer combo
        print "Binding to %s:%d" % pair
        self.socket.bind(pair)
        self.socket.listen(100)
        self.socket.setblocking(0)
        self.epoll = select.epoll()
        self.epoll.register(self.socket.fileno(), select.EPOLLIN)
                        
    def loop(self):
    
        e = {
            select.EPOLLIN:"EPOLLIN",
            select.EPOLLOUT:"EPOLLOUT", 
            select.EPOLLHUP:"EPOLLHUP",
            select.EPOLLERR:"EPOLLERR"}
        try:
            self.connections = {}
            
            while True:
                events = self.epoll.poll()
                for fileno, event in events:

                    if fileno == self.socket.fileno():
                        connection = Connection.objects.create(
                            self,
                            *self.socket.accept(),
                            reader=HttpRequestParser,
                            writer=HttpResponseRenderer)
                        print Colors.BOLD + Colors.BRIGHT_MAGENTA + "+++" + Colors.END,
                        print "Established connection (%d) from %s" % (connection.fd, connection.address),
                        if connection.country:
                            print "(%s)" % connection.country,
                        print "on port %d" % connection.port,

                        print "now polling %d connections" % len(self.connections)
                    else:

                        try:
                            connection = self.connections[fileno]
                        except KeyError:
                            print "!!! Warning: Received",
                            for code, name in e.items():
                                if event & code:
                                    print name,
                            print "for dropped socket %d" % fileno
                            sleep(0.25)
                            continue

                        if event & select.EPOLLIN:
                            # Pop bytes from the handle and try to process them
                            try:
                                connection.input.pop()
                            except Exception:
                                print traceback.format_exc()
                                connection.shutdown()
                                self.epoll.unregister(fileno)

                        elif event & select.EPOLLOUT:
                            try:
                                # Pop bytes from the buffer and write them to the socket
                                connection.output.pop()
                            except Exception:
                                print traceback.format_exc()
                                connection.shutdown()
                                self.epoll.unregister(fileno)
                        elif event & select.EPOLLHUP:
                            connection.forget("Connection hung up")
                            connection.handle.close()
                        elif event & select.EPOLLERR:
                            connection.forget("Connection error")
                            connection.handle.close()
                        else:
                            raise Exception("Unhandled event: %s" % e[event])

        except KeyboardInterrupt:
            print "Caught Ctrl-C..."
        finally:
            self.epoll.unregister(self.socket.fileno())
            self.epoll.close()
            self.socket.close()
           
    def route(self, **properties):
        # Route request
        try:
            
            controllers, properties = self.controller.route(**properties)
            while properties["path"].startswith("/"):
                properties["path"] = properties["path"][1:]
            # Create HTTP request instance from parsed data
            request = Request(**properties)
            
            # Process pre method calls        
            for controller in controllers:
                if hasattr(controller, "pre"):
                    controller.pre(request)

            # The last one is the dispatchable!
            endpoint = controllers[-1]
            
            # Call method
            response = endpoint.dispatch(request)
            
            # Bind request and response to eachother
            response.request, request.response = request, response
            response.headers["X-Served-By"] = "%s.%s" % (endpoint.__module__, endpoint.__class__.__name__)

            # Process post method call
            for controller in controllers:
                if hasattr(controller, "post"):
                    controller.post(request, response)

        except InvalidRoute:
            response = NotFoundResponse("Could not resolve request with path %s to any controller!" % properties["path"])
            response.disable_keepalive()
            
        properties["connection"].output.respond(response)


