
import re
from rusty.fields import *

class AbstractDatabase(object):
    def connect(self):
        raise NotImplementedError()
    
    def cursor(self):
        return self.connect().cursor()
        
    def batch(self, filename):
        cursor = self.cursor()
        with open(filename) as fh:
            for query in fh.read().split(";"):
                if query.strip():
                    print(re.sub("\n\s*", " ", query).strip())
                    cursor.execute(query)

class MysqlDatabase(AbstractDatabase):
    hostname = "127.0.0.1"
    port = 3306
    username = "root"
    password = "admin"
    database = "default"

    def __init__(self, hostname=None, port=None, username=None, password=None, database=None):
        if hostname:
            self.hostname = hostname
        if username:
            self.username = username
        if password:
            self.password = password
        if port:
            self.port = port
        if database:
            self.database = database
            
    def connect(self):
        import MySQLdb
        connection = MySQLdb.connect(
            host=self.hostname,
            user=self.username,
            passwd=self.password,
            port = self.port,
            db=self.database,
            use_unicode=True)
        connection.set_character_set('utf8')
        return connection
            
    def cursor(self):
        cursor = super(MysqlDatabase, self).cursor()
        cursor.execute('SET NAMES utf8;')
        cursor.execute('SET CHARACTER SET utf8;')
        cursor.execute('SET character_set_connection=utf8;')
        return cursor
        
class PostgreDatabase(AbstractDatabase):
    hostname = "127.0.0.1"
    port = 5432
    username = "postgres"
    password = "admin"
    database = "postgres"
    
    def __init__(self, hostname=None, port=None, username=None, password=None, database=None):
        if hostname:
            self.hostname = hostname
        if username:
            self.username = username
        if password:
            self.password = password
        if port:
            self.port = port
        if database:
            self.database = database
    
    def connect(self):
        import psycopg2
        
        connection = psycopg2.connect(
            host=self.hostname,
            user=self.username,
            password=self.password,
            port = self.port,
            database=self.database)
        
        return connection


class SqliteDatabase(AbstractDatabase):
    path = ":memory:"
    
    def __init__(self, path=None):
        if path:
            self.path = path
    
    def connect(self):
        import sqlite3
        return sqlite3.connect(self.path,
            detect_types=sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES,
            isolation_level=None)

class PostgreSchema(object):
    """
    Quick hack to generate PostgreSQL database schema from specified Rusty models
    """
    def __init__(self):
        self.tables = []
        self.m2m = []
        
    def map_field(self, model, field):
        if isinstance(field, DateTimeField):
            return "TIMESTAMP", "DEFAULT current_timestamp" if field.auto_now else None
        elif isinstance(field, UUIDField):
            return "UUID", "UNIQUE"
        elif isinstance(field, ForeignKey):
            datatype, extras = self.map_field(model, field.model._fields[field.model.objects.PRIMARY_KEY])
            return datatype, None
        elif isinstance(field, IntegerField):
            return "INTEGER", None
        elif isinstance(field, FloatField):
            return "FLOAT", None
        elif isinstance(field, BooleanField):
            return "BOOLEAN", None
        elif isinstance(field, ListField):
            self.add_m2m(field.model, model)
            return None, None
        elif isinstance(field, StringField):
            return "TEXT", None
        else:
            raise ValueError("Don't know how to map %s" % field)
        
    def add_model(self, *models):
        for model in models:
            model.prepare()
            table = "CREATE TABLE IF NOT EXISTS %s (\n" % model._get_name()
            table += "   id             SERIAL,\n"
            columns = []
            for name, field in sorted(model._fields.items()):#, key=lambda (name,field):field.null):
                datatype, extras = self.map_field(model, field)
                if not datatype:
                    continue
                column = "   "
                column += name.ljust(15)
                column += datatype
                if extras:
                    column += " " + extras
                if not field.null:
                    column += " NOT NULL"
                if name == model.objects.PRIMARY_KEY:
                    column += " PRIMARY KEY"
                columns.append(column)
            table += ",\n".join(columns)
            table += "\n);\n"

            self.tables.append(table)
        
    def add_m2m(self, left, right):
        table =  "CREATE TABLE IF NOT EXISTS %s_2_%s (\n" %  (left._get_name(), right._get_name())
        table += "    id SERIAL,\n"
        datatype, _ = self.map_field(left, left._fields[left.objects.PRIMARY_KEY])
        table += "    %s_%s %s,\n" % (left._get_name(), left.objects.PRIMARY_KEY, datatype)
        datatype, _ = self.map_field(right, right._fields[right.objects.PRIMARY_KEY])
        table += "    %s_%s %s\n" % (right._get_name(), right.objects.PRIMARY_KEY, datatype)
        table += ");"
        self.m2m.append(table)

    def dump(self):
        return "\n".join(self.tables + self.m2m)
