
import pwd, grp, os, types
from rusty.jason import json_decode, json_encode
from tempfile import NamedTemporaryFile
from pwd import getpwnam
from threading import Lock, Timer

class FlatFileStorage(object):
    def __init__(self, filename, write_timeout=0):
        self.write_timeout = write_timeout
        self.timer = None
        self.lock = Lock()
        self.filename = os.path.realpath(os.path.expanduser(filename))
        if os.path.exists(self.filename):
            with open(self.filename) as fh:
                self.data = json_decode(fh.read())
                assert isinstance(self.data, types.DictType)
        else:
            self.data = {}
            
    def __repr__(self):
        return "<FlatFileStorage at %s>" % self.filename
        
    def retrieve(self, key):
        with self.lock:
            return self.data[key]
            
    def _write(self):
        print "Writing:", self.filename
        fh = NamedTemporaryFile(
            suffix=".json~",
            dir=os.path.dirname(self.filename),
            delete=False)
        fh.write(json_encode(self.data))
        fh.close()
        os.rename(fh.name, self.filename)
        
    def write(self):
        with self.lock:
            self._write()

    def replace(self, key, value):
        with self.lock:
            self.data[key] = value
            if self.write_timeout:
                try:
                    self.timer.cancel()
                except AttributeError:
                    pass
                self.timer = Timer(self.write_timeout, self._write)
                self.timer.start()
            else:
                self._write()
    
    def keys(self):
        return self.data.keys()
        
    def __iter__(self):
        return self.data.iterkeys()
        
    def list(self):
        return self.data.keys()
    
class MultiFileStorage(object):
    """
    Store each model instance in a separate file. The primary key is suffixed
    with .json to form the filename.
    """
    def __init__(self, root, user="koodur", group="koodur"):
        self.uid = pwd.getpwnam(user).pw_uid if user else None
        self.gid = grp.getgrnam(group).gr_gid if group else None

        assert isinstance(root, basestring)

        self.root = os.path.realpath(os.path.expanduser(root))

        self.lock = Lock()
        if not os.path.exists(self.root):
            os.makedirs(self.root)
            
    def __repr__(self):
        return "<MultiFileStorage at %s>" % self.root

    def resolve_path(self, key):
        return os.path.join(self.root, "%s.json" % key)
        
    def __iter__(self):
        for filename in os.listdir(self.root):
            if filename.endswith(".json"):
                yield filename[:-5]
                
    def list(self):
        return tuple(iter(self))
        
    def keys(self):
        return tuple(self)

    def replace(self, key, value):
        # Atomicity guaranteed with lock and renaming
        with self.lock:
            # Don't allow slashes for directory tree traversal
            key = str(key)
            assert "/" not in key and "\0" not in key, "Invalid primary key %s" % key
            assert isinstance(value, dict)
            fh = NamedTemporaryFile(
                suffix=".json~",
                dir=self.root,
                delete=False)
            fh.write(json_encode(value))
            fh.close()
            
            # In case we're running as root at this point
            os.chown(fh.name, self.uid, self.gid)
            
            # Rename temporary file
            os.rename(fh.name, self.resolve_path(key))
            
    def retrieve(self, key):
        assert "/" not in key and "\0" not in key, "Invalid primary key %s" % key
        with self.lock:
            path = self.resolve_path(key)
            assert os.path.exists(path), "Key %s does not exist" % key
            with open(path, "r") as fh:
                return json_decode(fh.read())

