
import sys
import os
import codecs
from rusty.pipeline import PipelineMixin

class Stream(PipelineMixin):
    def __init__(self, *args, **kwargs):
        raise NotImplementedError('Stream class constructor not overridden')

    def close(self):
        self._handle.close()

    def closed(self):
        return self._handle.fileno == None

    def fileno(self):
        return self._handle.fileno()

    def isatty(self):
        return False
        
    def seek(offset, whence=0):
        raise NotImplementedError('seek() not implemented')
        
    def seekable(self):
        raise NotImplementedError('seekable() method not implemented')
        
    def tell(self):
        return self._handle.tell()
        
    def __repr__(self):
        return "<Stream>"
        
class InputStream(Stream):
    def __init__(self, handle, filename=None, mimetype=None):
        self._handle = handle
        self.filename = filename
        self.mimetype = mimetype
        
    def __repr__(self):
        return "<InputStream>"
        
    def flush(self):
        pass
        
    def isatty(self):
        return False

    def __iter__(self, chunk_size=32768):
        while True:
            chunk = self.read(chunk_size)
            if not chunk:
                break
            yield chunk
        self.close()

    def read(self, size=-1):
        return self._handle.read(size)

    def readable(self):
        return True
        
    def readline(self, limit=-1):
        return self._handle.readline(limit)
        
    def readlines(self, hint=0):
        return self._handle.readlines(hint)

    def truncate(self, size=None):
        raise IOError('Unable to truncate read-only stream')
        
    def writable(self):
        return False
        
    def writelines(self, lines):
        raise IOError('Unable to write to read-only stream')
        
class OutputStream(Stream):
    def __repr__(self):
        return "<OutputStream>"
        
    def pipe(self, haystack):
        for needle in haystack:
            self.write(needle)
            yield needle
        
    def flush(self):
        self._handle.flush()
        
    def readable(self):
        return False
        
    def readline(self, limit=-1):
        raise IOError('readline() method not applicable for %s' % self.__class__)

    def seekable(self):
        return False
        
    def truncate(self, size=None):
        self._handle.truncate(size)

    def writable(self):
        return True
        
    def write(self, buf):
        self._handle.write(buf)
    
    def writelines(self, lines):
        self._handle.writelines(lines)

class InputFileStream(InputStream):
    def __init__(self, path):
        self.path = path
        self._handle = open(path, 'rb')
        
    def seekable(self):
        return True
        
    def __repr__(self):
        return "<InputFileStream of %s>" % self.path
        
            
class Buffer(PipelineMixin):
    def __init__(self, buf):
        self._buffer = buf
        
    def __iter__(self):
        yield self._buffer
        
class Connection(Stream):
    """
    Input/output stream bound to a socket connection
    """
    def __init__(self, connection, address=None):
        self._handle = connection
        connection.setblocking(0)
        self.address = address

    def read(self, limit):
        return self._handle.recv(limit)
        
    def write(self, buf):
        # This is quite naive code, some buffers might get lost!
        self._handle.sendall(buf)

    def __repr__(self):
        if self.address:
            return "<Connection from %s:%d>" % self.address
        else:
            return "<Connection>"
        
class InputTextFileStream(InputFileStream):
    def __init__(self, path):
        self.path = path
        self._handle = codecs.open(path, 'r', encoding="utf-8")

    def __repr__(self):
        return "<InputTextFileStream of %s>" % self.path
        
    def __iter__(self):
        def generator():
            while True:
                line = self._handle.readline()
                if line:
                    yield line.strip()
                else:
                    break
            self.close()
        return generator()

class OutputFileStream(OutputStream):
    def __init__(self, path, overwrite=False):
        self.path = path

        if not overwrite and os.path.exists(path):
            raise IOError('Will not overwrite %s' % path)
            
        self._handle = open(path, 'wb')
    
    def __repr__(self):
        return "<OutputFileStream of %s>" % self.path
                    
class StandardOutputStream(OutputStream):
    def __init__(self):
        self._handle = sys.stdout

