
from jinja2 import Markup, Template, Environment, FileSystemLoader
from docutils.core import publish_parts
from docutils import nodes
from docutils.parsers.rst import directives, Directive
from pygments import highlight
from pygments.lexers import get_lexer_by_name, TextLexer
from pygments.formatters import HtmlFormatter
from rusty.fields import Field
from rusty.http import Response

class Pygments(Directive):
    """
    Source code syntax hightlighting.
    """
    required_arguments = 1
    optional_arguments = 1
    final_argument_whitespace = True
    has_content = True
    linenos=1

    def run(self):
        dic = dict([options.split("=") for option in self.arguments[1:]])
        self.assert_has_content()
        try:
            lexer = get_lexer_by_name(self.arguments[0], **dic)
        except ValueError:
            lexer = TextLexer()
        formatter = HtmlFormatter(
            noclasses=False,
            linenos="table",
            cssclass="codehilite")
        parsed = highlight(u'\n'.join(self.content), lexer, formatter)
        return [nodes.raw('', parsed, format="html")]


directives.register_directive('code', Pygments)

def naturalsize(value, binary=False, gnu=False):
    """Format a number of byteslike a human readable filesize (eg. 10 kB).  By
    default, decimal suffixes (kB, MB) are used.  Passing binary=true will use
    binary suffixes (KiB, MiB) are used and the base will be 2**10 instead of
    10**3.  If ``gnu`` is True, the binary argument is ignored and GNU-style
    (ls -sh style) prefixes are used (K, M) with the 2**10 definition.
    Non-gnu modes are compatible with jinja2's ``filesizeformat`` filter."""
    if gnu: suffix = "KMGTPEZY"
    elif binary: suffix = 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB'
    else: suffix = 'kB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'

    base = 1024 if (gnu or binary) else 1000
    bytes = float(value)

    if bytes == 1 and not gnu: return '1 Byte'
    elif bytes < base and not gnu: return '%d Bytes' % bytes
    elif bytes < base and gnu: return '%dB' % bytes

    for i,s in enumerate(suffix):
        unit = base ** (i+2)
        if bytes < unit and not gnu:
            return '%.1f %s' % ((base * bytes / unit), s)
        elif bytes < unit and gnu:
            return '%.1f%s' % ((base * bytes / unit), s)
    if gnu:
        return '%.1f%s' % ((base * bytes / unit), s)
    return '%.1f %s' % ((base * bytes / unit), s)


class JinjaResponse(Response):
    def render(self, template, context):
        template = self.request.application.jinja_environment.get_template(template)        
        yield template.render(context).encode("utf-8")

    def __init__(self, template="index.html", environment=None, status=200, context={}):
        super(JinjaResponse, self).__init__(
            body=self.render(template, context),
            status=status,
            headers={
                "Content-Type":"text/html; charset=utf-8",
                "Content-Encoding":"chunked"
            })

class JinjaEnvironmentField(Field):
    def parse(self, mixed):
        if isinstance(mixed, basestring):
            mixed = Environment(loader=FileSystemLoader(mixed))
            mixed.filters["naturalsize"] = naturalsize
            mixed.filters["rst"] = lambda s: publish_parts(
                source=s,
                writer_name="html",
                settings_overrides={"report_level":"quiet"})["body"]
        if isinstance(mixed, Environment):
            return mixed
        raise ValueError("Don't know how to interpret %s as Jinja environment" % mixed)
