
import array
import struct
import types
from rusty.signals import Signal
from rusty.events import Event
from rusty.jason import json_encode, json_decode, JasonDecodeError

class WebSocketFrame(object):
    def __init__(self, opcode, payload, masked=False):
        self.final = True
        self.masked = masked
        self.opcode = opcode
        self.payload = payload
        
    def render(self):
        if self.final:
            final_bit = 0x80
        else:
            final_bit = 0
        header = struct.pack("B", final_bit | self.opcode)
        
        payload = self.payload
        
        # Encode Unicode payload
        if isinstance(payload, types.UnicodeType):
            payload = payload.encode("utf-8")
        payload_length = len(self.payload)
        
        if self.masked:
            mask_bit = 0x80
        else:
            mask_bit = 0
        if payload_length < 126:
            header += struct.pack("B", payload_length | mask_bit)
        elif payload_length <= 0xFFFF:
            header += struct.pack("!BH", 126 | mask_bit, payload_length)
        else:
            header += struct.pack("!BQ", 127 | mask_bit, payload_length)
        
        if self.masked:
            mask = os.urandom(4)
            header += mask
            
            mask = array.array("B", mask)
            
            payload = array.array("B", payload)
            for i in xrange(payload_length):
                payload[i] = payload[i] ^ mask[i % 4]
            payload = payload.tostring()
            
        return header + payload
        
class WebSocketParser(object):
    def __init__(self, connection, message_received=None):
        self.connection = connection
        self.buffer = ""
        self.message_received = message_received or Signal()
        self.message = "" # TODO: Multiplexing ?!
        
    def __repr__(self):
        return "<WebSocketParser %d bytes buffered>" % len(self.buffer)
    
    def pop(self, buffer_size=32768):
        """
        Pop some bytes from incoming socket and try to parse them
        """
        chunk = self.connection.recv(buffer_size)
        
        if not chunk:
            # We did get no bytes, assume client disconnected
            self.connection.forget("WebSocket did not receive any bytes")
            return
            
        self.buffer += chunk
        if len(self.buffer) < 2: # No frames shorter than 2 bytes!
            return
        
        payload_offset = 2
        mask_offset = 2
        
        header, payload_length = struct.unpack("BB", self.buffer[:2])

        # Extract header
        final_frame = header & 0x80
        reserved_bits = header & 0x70
        frame_opcode = header & 0xf
        frame_opcode_is_control = frame_opcode & 0x8
        
        if reserved_bits:
            raise Exception("Don't know how to handle reserved bits")
            
        # Extract payload length
        masked_frame = bool(payload_length & 0x80)
        payload_length = payload_length & 0x7f
        
        if masked_frame:
            payload_offset += 4
        
        if frame_opcode_is_control and payload_length >= 126:
            raise Exception("Control frames must have payload < 126")

        if payload_length == 126:
            if len(self.buffer) < 4: # Not enough bytes in this iteration
                return
            payload_offset += 2
            mask_offset += 2
            payload_length, = struct.unpack("!H", self.buffer[2:4])
        elif payload_length == 127:
            if len(self.buffer) < 10: # Not enough bytes this time!
                return
            payload_offset += 8
            mask_offset += 8
            payload_length, = struct.unpack("!Q", self.buffer[2:10])

        # Extract payload portion from buffer
        if len(self.buffer) < payload_offset+payload_length:
            return
        payload = self.buffer[payload_offset:payload_offset+payload_length]
        
        # Unmask if necessary
        if masked_frame:
            mask = array.array("B", self.buffer[mask_offset:mask_offset+4])
            payload = array.array("B", payload)
            for i in xrange(payload_length):
                payload[i] = payload[i] ^ mask[i % 4]
            payload = payload.tostring()

        # Append to message buffer
        self.message += payload
        self.buffer = self.buffer[payload_offset+payload_length:]
                
        # If this is the final frame
        if final_frame:
            try:
                properties = json_decode(self.message)
            except JasonDecodeError:
                print "Failed to parse JSON from %s" % self.message
            else:
                properties["user"] = self.connection.user
                assert self.connection.uuid, "Invalid connection UUID: %s" % self.connection.uuid
                properties["origin"] = self.connection.uuid
                Event.objects.create(**properties)
            finally:
                self.message = ""
                


class WebSocketRenderer(object):
    def __init__(self, connection):
        self.connection = connection
        self.buffer = ""
        Event.objects.created.connect(self.publish)
        
    def __repr__(self):
        return "<WebSocketRenderer %d bytes remaining>" % len(self.buffer)
        
    def publish(self, event):
        # Avoid cyclic event forwarding
        if event.origin == self.connection.uuid:
            return
        properties = event.serialize(namespace=True)
        properties["name"] = event._get_name()
        self.send_message(json_encode(properties))
        
    def send_message(self, message):
        self.send(WebSocketFrame(0x01, message))
        
    def send(self, frame):
        """
        Append rendered frame to the buffer
        TODO: Implement queue of frames
        """
        self.buffer += frame.render()
        self.connection.set_duplex()
    
    def pop(self, buffer_size=32768):
        """
        Pop some bytes from the output buffer and write them to the
        outbound socket
        """
        if self.buffer:
            bytes_written = self.connection.send(self.buffer)
            self.buffer = self.buffer[bytes_written:]
        else:
            bytes_written = 0

        if not bytes_written:
            self.connection.set_incoming()
